"""A cmd2 script that saves the help text for every command, subcommand, and topic to a file.
This is meant to be run within a cmd2 session using run_pyscript.
"""

import argparse
import os
import sys
from typing import TextIO

ASTERISKS = "********************************************************"


def get_sub_commands(parser: argparse.ArgumentParser) -> list[str]:
    """Get a list of subcommands for an ArgumentParser."""
    sub_cmds = []

    # Check if this is parser has subcommands
    if parser is not None and parser._subparsers is not None:
        # Find the _SubParsersAction for the subcommands of this parser
        for action in parser._subparsers._actions:
            if isinstance(action, argparse._SubParsersAction):
                for sub_cmd, sub_cmd_parser in action.choices.items():
                    sub_cmds.append(sub_cmd)

                    # Look for nested subcommands
                    sub_cmds.extend(f'{sub_cmd} {nested_sub_cmd}' for nested_sub_cmd in get_sub_commands(sub_cmd_parser))
                break

    sub_cmds.sort()
    return sub_cmds


def add_help_to_file(item: str, outfile: TextIO, is_command: bool) -> None:
    """Write help text for commands and topics to the output file
    :param item: what is having its help text saved
    :param outfile: file being written to
    :param is_command: tells if the item is a command and not just a help topic.
    """
    label = "COMMAND" if is_command else "TOPIC"

    header = f'{ASTERISKS}\n{label}: {item}\n{ASTERISKS}\n'
    outfile.write(header)

    result = app(f'help {item}')
    outfile.write(result.stdout)


def main() -> None:
    """Main function of this script."""
    # Make sure we have access to self
    if 'self' not in globals():
        print("Re-run this script from a cmd2 application where self_in_py is True")
        return

    # Make sure the user passed in an output file
    if len(sys.argv) != 2:
        print(f"Usage: {os.path.basename(sys.argv[0])} <output_file>")
        return

    # Open the output file
    outfile_path = os.path.expanduser(sys.argv[1])
    try:
        with open(outfile_path, 'w') as outfile:
            pass
    except OSError as e:
        print(f"Error opening {outfile_path} because: {e}")
        return

    # Write the help summary
    header = f'{ASTERISKS}\nSUMMARY\n{ASTERISKS}\n'
    outfile.write(header)

    result = app('help -v')
    outfile.write(result.stdout)

    # Get a list of all commands and help topics and then filter out duplicates
    all_commands = set(self.get_all_commands())
    all_topics = set(self.get_help_topics())
    to_save = list(all_commands | all_topics)
    to_save.sort()

    for item in to_save:
        is_command = item in all_commands
        add_help_to_file(item, outfile, is_command)

        if is_command:
            # Add any subcommands
            for subcmd in get_sub_commands(getattr(self.cmd_func(item), 'argparser', None)):
                full_cmd = f'{item} {subcmd}'
                add_help_to_file(full_cmd, outfile, is_command)

    outfile.close()
    print(f"Output written to {outfile_path}")


# Run main function
main()



"""Unit testing for cmd2/terminal_utils.py module"""

import pytest

from cmd2 import (
    Color,
)
from cmd2 import string_utils as su
from cmd2 import terminal_utils as tu


def test_set_title() -> None:
    title = "Hello, world!"
    assert tu.set_title_str(title) == tu.OSC + '2;' + title + tu.BEL


@pytest.mark.parametrize(
    ('cols', 'prompt', 'line', 'cursor', 'msg', 'expected'),
    [
        (
            127,
            '(Cmd) ',
            'help his',
            12,
            su.stylize('Hello World!', style=Color.MAGENTA),
            '\x1b[2K\r\x1b[35mHello World!\x1b[0m',
        ),
        (127, '\n(Cmd) ', 'help ', 5, 'foo', '\x1b[2K\x1b[1A\x1b[2K\rfoo'),
        (
            10,
            '(Cmd) ',
            'help history of the american republic',
            4,
            'boo',
            '\x1b[3B\x1b[2K\x1b[1A\x1b[2K\x1b[1A\x1b[2K\x1b[1A\x1b[2K\x1b[1A\x1b[2K\rboo',
        ),
    ],
)
def test_async_alert_str(cols, prompt, line, cursor, msg, expected) -> None:
    alert_str = tu.async_alert_str(terminal_columns=cols, prompt=prompt, line=line, cursor_offset=cursor, alert_msg=msg)
    assert alert_str == expected


def test_clear_screen() -> None:
    clear_type = 2
    assert tu.clear_screen_str(clear_type) == f"{tu.CSI}{clear_type}J"

    clear_type = -1
    expected_err = "clear_type must in an integer from 0 to 3"
    with pytest.raises(ValueError, match=expected_err):
        tu.clear_screen_str(clear_type)

    clear_type = 4
    with pytest.raises(ValueError, match=expected_err):
        tu.clear_screen_str(clear_type)


def test_clear_line() -> None:
    clear_type = 2
    assert tu.clear_line_str(clear_type) == f"{tu.CSI}{clear_type}K"

    clear_type = -1
    expected_err = "clear_type must in an integer from 0 to 2"
    with pytest.raises(ValueError, match=expected_err):
        tu.clear_line_str(clear_type)

    clear_type = 3
    with pytest.raises(ValueError, match=expected_err):
        tu.clear_line_str(clear_type)


def test_cursor() -> None:
    count = 1
    assert tu.Cursor.UP(count) == f"{tu.CSI}{count}A"
    assert tu.Cursor.DOWN(count) == f"{tu.CSI}{count}B"
    assert tu.Cursor.FORWARD(count) == f"{tu.CSI}{count}C"
    assert tu.Cursor.BACK(count) == f"{tu.CSI}{count}D"

    x = 4
    y = 5
    assert tu.Cursor.SET_POS(x, y) == f"{tu.CSI}{y};{x}H"



#!/usr/bin/env python
"""A sample application for how Python scripting can provide conditional
control flow of a cmd2 application.

cmd2's built-in scripting capability, which can be invoked via the "@" shortcut
or "run_script" command, uses basic ASCII/UTF-8 text scripts and is very easy
to use.  Moreover, the trivial syntax of the script files, where there is one
command per line and the line is exactly what the user would type inside the
application, makes it so non-technical that end users can quickly learn to
create scripts.

However, there comes a time when technical end users want more capability and
power.  In particular it is common that users will want to create a script with
conditional control flow - where the next command run will depend on the
results from the previous command.  This is where the ability to run Python
scripts inside a cmd2 application via the run_pyscript command and the
"run_pyscript <script> [arguments]" syntax comes into play.

This application and the "examples/scripts/conditional.py" script serve as an
example for one way in which this can be done.
"""

import argparse
import os

import cmd2
from cmd2 import (
    Color,
    stylize,
)


class CmdLineApp(cmd2.Cmd):
    """Example cmd2 application to showcase conditional control flow in Python scripting within cmd2 apps."""

    def __init__(self) -> None:
        # Set include_ipy to True to enable the "ipy" command which runs an interactive IPython shell
        super().__init__(include_ipy=True)
        self._set_prompt()
        self.intro = 'Happy 𝛑 Day.  Note the full Unicode support:  😇 💩'

    def _set_prompt(self) -> None:
        """Set prompt so it displays the current working directory."""
        self.cwd = os.getcwd()
        self.prompt = stylize(f'{self.cwd} $ ', style=Color.CYAN)

    def postcmd(self, stop: bool, _line: str) -> bool:
        """Hook method executed just after a command dispatch is finished.

        :param stop: if True, the command has indicated the application should exit
        :param line: the command line text for this command
        :return: if this is True, the application will exit after this command and the postloop() will run
        """
        """Override this so prompt always displays cwd."""
        self._set_prompt()
        return stop

    @cmd2.with_argument_list
    def do_cd(self, arglist: list[str]) -> None:
        """Change directory.
        Usage:
            cd <new_dir>.
        """
        # Expect 1 argument, the directory to change to
        if not arglist or len(arglist) != 1:
            self.perror("cd requires exactly 1 argument:")
            self.do_help('cd')
            self.last_result = 'Bad arguments'
            return

        # Convert relative paths to absolute paths
        path = os.path.abspath(os.path.expanduser(arglist[0]))

        # Make sure the directory exists, is a directory, and we have read access
        err = None
        data = None
        if not os.path.isdir(path):
            err = f'{path} is not a directory'
        elif not os.access(path, os.R_OK):
            err = f'You do not have read access to {path}'
        else:
            try:
                os.chdir(path)
            except Exception as ex:  # noqa: BLE001
                err = f'{ex}'
            else:
                self.poutput(f'Successfully changed directory to {path}')
                data = path

        if err:
            self.perror(err)
        self.last_result = data

    # Enable tab completion for cd command
    def complete_cd(self, text: str, line: str, begidx: int, endidx: int) -> list[str]:
        # Tab complete only directories
        return self.path_complete(text, line, begidx, endidx, path_filter=os.path.isdir)

    dir_parser = cmd2.Cmd2ArgumentParser()
    dir_parser.add_argument('-l', '--long', action='store_true', help="display in long format with one item per line")

    @cmd2.with_argparser(dir_parser, with_unknown_args=True)
    def do_dir(self, _args: argparse.Namespace, unknown: list[str]) -> None:
        """List contents of current directory."""
        # No arguments for this command
        if unknown:
            self.perror("dir does not take any positional arguments:")
            self.do_help('dir')
            self.last_result = 'Bad arguments'
            return

        # Get the contents as a list
        contents = os.listdir(self.cwd)

        for f in contents:
            self.poutput(f'{f}')
        self.poutput('')

        self.last_result = contents


if __name__ == '__main__':
    import sys

    c = CmdLineApp()
    sys.exit(c.cmdloop())



#!/usr/bin/env python3
"""An example of using Rich Tables within a cmd2 application for displaying tabular data.

While you can use any Python library for displaying tabular data within a cmd2 application,
we recommend using rich since that is built into cmd2.

Data comes from World Population Review: https://worldpopulationreview.com/
and https://en.wikipedia.org/wiki/List_of_countries_by_GDP_(nominal)

NOTE: The flag emojis used require appropriate font support in your terminal
and/or IDE. Any "Nerd Font" should work. Some good options include:
- JetBrains Mono
- Fira Code
- Hack
- Monaspace
"""

from rich.table import Table

import cmd2
from cmd2.colors import Color

CITY_HEADERS = ['Flag', 'City', 'Country', '2025 Population']
CITY_DATA = [
    ["🇯🇵", "Tokyo (東京)", "Japan", 37_036_200],
    ["🇮🇳", "Delhi", "India", 34_665_600],
    ["🇨🇳", "Shanghai (上海)", "China", 30_482_100],
    ["🇧🇩", "Dhaka", "Bangladesh", 24_652_900],
    ["🇪🇬", "Cairo (القاهرة)", "Egypt", 23_074_200],
    ["🇪🇬", "São Paulo", "Brazil", 22_990_000],
    ["🇲🇽", "Mexico City", "Mexico", 22_752_400],
    ["🇨🇳", "Beijing (北京)", "China", 22_596_500],
    ["🇮🇳", "Mumbai", "India", 22_089_000],
    ["🇯🇵", "Osaka (大阪)", "Japan", 18_921_600],
]
CITY_TITLE = "10 Largest Cities by Population 2025"
CITY_CAPTION = "Data from https://worldpopulationreview.com/"

COUNTRY_HEADERS = [
    'Flag',
    'Country',
    '2025 Population',
    'Area (M km^2)',
    'Population Density (/km^2)',
    'GDP (million US$)',
    'GDP per capita (US$)',
]
COUNTRY_DATA = [
    ["🇮🇳", "India", 1_463_870_000, 3.3, 492, 4_187_017, 2_878],
    ["🇨🇳", "China (中国)", 1_416_100_000, 9.7, 150, 19_231_705, 13_687],
    ["🇺🇸", "United States", 347_276_000, 9.4, 38, 30_507_217, 89_105],
    ["🇮🇩", "Indonesia", 285_721_000, 1.9, 152, 1_429_743, 5_027],
    ["🇵🇰", "Pakistan", 255_220_000, 0.9, 331, 373_072, 1_484],
    ["🇳🇬", "Nigeria", 237_528_000, 0.9, 261, 188_271, 807],
    ["🇧🇷", "Brazil", 212_812_000, 8.5, 25, 2_125_958, 9_964],
    ["🇧🇩", "Bangladesh", 175_687_000, 0.1, 1_350, 467_218, 2_689],
    ["🇷🇺", "Russia (россия)", 143_997_000, 17.1, 9, 2_076_396, 14_258],
    ["🇪🇹", "Ethiopia (እትዮጵያ)", 135_472_000, 1.1, 120, 117_457, 1_066],
]
COUNTRY_TITLE = "10 Largest Countries by Population 2025"
COUNTRY_CAPTION = "Data from https://worldpopulationreview.com/ and Wikipedia"


class TableApp(cmd2.Cmd):
    """Cmd2 application to demonstrate displaying tabular data using rich."""

    TABLE_CATEGORY = 'Table Commands'

    def __init__(self) -> None:
        """Initialize the cmd2 application."""
        super().__init__()

        # Prints an intro banner once upon application startup
        self.intro = 'Are you curious which countries and cities on Earth have the largest populations?'

        # Set the default category name
        self.default_category = 'cmd2 Built-in Commands'

    @cmd2.with_category(TABLE_CATEGORY)
    def do_cities(self, _: cmd2.Statement) -> None:
        """Display the cities with the largest population."""
        table = Table(title=CITY_TITLE, caption=CITY_CAPTION)

        for header in CITY_HEADERS:
            table.add_column(header)

        for row in CITY_DATA:
            # Convert integers or floats to strings, since rich tables can not render int/float
            str_row = [f"{item:,}" if isinstance(item, int) else str(item) for item in row]
            table.add_row(*str_row)

        self.poutput(table)

    @cmd2.with_category(TABLE_CATEGORY)
    def do_countries(self, _: cmd2.Statement) -> None:
        """Display the countries with the largest population."""
        table = Table(title=COUNTRY_TITLE, caption=COUNTRY_CAPTION)

        for header in COUNTRY_HEADERS:
            justify = "right"
            header_style = None
            style = None
            match header:
                case population if "2025 Population" in population:
                    header_style = Color.BRIGHT_BLUE
                    style = Color.BLUE
                case density if "Density" in density:
                    header_style = Color.BRIGHT_RED
                    style = Color.RED
                case percap if "per capita" in percap:
                    header_style = Color.BRIGHT_GREEN
                    style = Color.GREEN
                case flag if 'Flag' in flag:
                    justify = "center"
                case country if 'Country' in country:
                    justify = "left"

            table.add_column(header, justify=justify, header_style=header_style, style=style)

        for row in COUNTRY_DATA:
            # Convert integers or floats to strings, since rich tables can not render int/float
            str_row = [f"{item:,}" if isinstance(item, int) else str(item) for item in row]
            table.add_row(*str_row)

        self.poutput(table)


if __name__ == '__main__':
    app = TableApp()
    app.cmdloop()



"""This is a Python script intended to be used with the "python_scripting.py" cmd2 example application.

To run it you should do the following:
    ./python_scripting.py
    run_pyscript scripts/conditional.py directory_path

Note: The "app" function is defined within the cmd2 embedded Python environment and in there "self" is your cmd2
application instance. Note: self only exists in this environment if self_in_py is True.
"""

import os
import sys

if len(sys.argv) > 1:
    directory = sys.argv[1]
    print(f'Using specified directory: {directory!r}')
else:
    directory = 'foobar'
    print(f'Using default directory: {directory!r}')

# Keep track of where we stared
original_dir = os.getcwd()

# Try to change to the specified directory
result = app(f'cd {directory}')

# Conditionally do something based on the results of the last command
if result:
    print(f"STDOUT: {result.stdout}\n")
    print(f"STDERR: {result.stderr}\n")

    print(f'\nContents of directory {directory!r}:')
    result = app('dir -l')

    print(f"STDOUT: {result.stdout}\n")
    print(f"STDERR: {result.stderr}\n")

    print(f'{result.data}\n')

    # Change back to where we were
    print(f'Changing back to original directory: {original_dir!r}')
    app(f'cd {original_dir}')
else:
    # cd command failed, print a warning
    print(f'Failed to change directory to {directory!r}')

    print(f"STDOUT: {result.stdout}\n")
    print(f"STDERR: {result.stderr}\n")



#!/usr/bin/env python
"""An example cmd2 mixin."""

import functools
from collections.abc import Callable
from typing import TYPE_CHECKING

import cmd2

if TYPE_CHECKING:  # pragma: no cover
    _Base = cmd2.Cmd
else:
    _Base = object


def empty_decorator(func: Callable) -> Callable:
    """An empty decorator for use with your mixin."""

    @functools.wraps(func)
    def _empty_decorator(self, *args, **kwargs) -> None:
        self.poutput("in the empty decorator")
        func(self, *args, **kwargs)

    _empty_decorator.__doc__ = func.__doc__
    return _empty_decorator


class MyMixin(_Base):
    """A mixin class which adds a 'say' command to a cmd2 subclass.

    The order in which you add the mixin matters. Say you want to
    use this mixin in a class called MyApp.

    class MyApp(MyMixin, cmd2.Cmd):
        def __init__(self, *args, **kwargs):
            # gotta have this or neither the mixin or cmd2 will initialize
            super().__init__(*args, **kwargs)
    """

    def __init__(self, *args, **kwargs) -> None:
        # code placed here runs before cmd2 initializes
        super().__init__(*args, **kwargs)
        # code placed here runs after cmd2 initializes
        # this is where you register any hook functions
        self.register_preloop_hook(self.cmd2_mymixin_preloop_hook)
        self.register_postloop_hook(self.cmd2_mymixin_postloop_hook)
        self.register_postparsing_hook(self.cmd2_mymixin_postparsing_hook)

    def do_say(self, statement) -> None:
        """Simple say command."""
        self.poutput(statement)

    #
    # define hooks as functions, not methods
    def cmd2_mymixin_preloop_hook(self) -> None:
        """Method to be called before the command loop begins."""
        self.poutput("preloop hook")

    def cmd2_mymixin_postloop_hook(self) -> None:
        """Method to be called after the command loop finishes."""
        self.poutput("postloop hook")

    def cmd2_mymixin_postparsing_hook(self, data: cmd2.plugin.PostparsingData) -> cmd2.plugin.PostparsingData:
        """Method to be called after parsing user input, but before running the command."""
        self.poutput('in postparsing hook')
        return data


class Example(MyMixin, cmd2.Cmd):
    """An class to show how to use a mixin."""

    def __init__(self, *args, **kwargs) -> None:
        # gotta have this or neither the mixin or cmd2 will initialize
        super().__init__(*args, **kwargs)

    @empty_decorator
    def do_something(self, _arg) -> None:
        self.poutput('this is the something command')


if __name__ == '__main__':
    app = Example()
    app.cmdloop()



"""Unit testing for get_defining_class in cmd2/utils.py module."""

import functools

import cmd2.utils as cu


class ParentClass:
    def func_with_overrides(self) -> None:
        pass

    def parent_only_func(self, param1, param2) -> None:
        pass


class ChildClass(ParentClass):
    def func_with_overrides(self) -> None:
        super().func_with_overrides()

    def child_function(self) -> None:
        pass

    def lambda1() -> int:
        return 1

    def lambda2() -> int:
        return 2

    @classmethod
    def class_method(cls) -> None:
        pass

    @staticmethod
    def static_meth() -> None:
        pass


def func_not_in_class() -> None:
    pass


def test_get_defining_class() -> None:
    parent_instance = ParentClass()
    child_instance = ChildClass()

    # validate unbound class functions
    assert cu.get_defining_class(ParentClass.func_with_overrides) is ParentClass
    assert cu.get_defining_class(ParentClass.parent_only_func) is ParentClass
    assert cu.get_defining_class(ChildClass.func_with_overrides) is ChildClass
    assert cu.get_defining_class(ChildClass.parent_only_func) is ParentClass
    assert cu.get_defining_class(ChildClass.child_function) is ChildClass
    assert cu.get_defining_class(ChildClass.class_method) is ChildClass
    assert cu.get_defining_class(ChildClass.static_meth) is ChildClass

    # validate bound class methods
    assert cu.get_defining_class(parent_instance.func_with_overrides) is ParentClass
    assert cu.get_defining_class(parent_instance.parent_only_func) is ParentClass
    assert cu.get_defining_class(child_instance.func_with_overrides) is ChildClass
    assert cu.get_defining_class(child_instance.parent_only_func) is ParentClass
    assert cu.get_defining_class(child_instance.child_function) is ChildClass
    assert cu.get_defining_class(child_instance.class_method) is ChildClass
    assert cu.get_defining_class(child_instance.static_meth) is ChildClass

    # bare functions resolve to nothing
    assert cu.get_defining_class(func_not_in_class) is None

    # lambdas and nested lambdas
    assert cu.get_defining_class(ChildClass.lambda1) is ChildClass
    assert cu.get_defining_class(ChildClass.lambda2) is ChildClass
    assert cu.get_defining_class(ChildClass().lambda1) is ChildClass
    assert cu.get_defining_class(ChildClass().lambda2) is ChildClass

    # partials
    partial_unbound = functools.partial(ParentClass.parent_only_func, 1)
    nested_partial_unbound = functools.partial(partial_unbound, 2)
    assert cu.get_defining_class(partial_unbound) is ParentClass
    assert cu.get_defining_class(nested_partial_unbound) is ParentClass

    partial_bound = functools.partial(parent_instance.parent_only_func, 1)
    nested_partial_bound = functools.partial(partial_bound, 2)
    assert cu.get_defining_class(partial_bound) is ParentClass
    assert cu.get_defining_class(nested_partial_bound) is ParentClass



#!/usr/bin/env python
"""A sample application for integrating cmd2 with external event loops.

This is an example of how to use cmd2 in a way so that cmd2 doesn't own the inner event loop of your application.

This opens up the possibility of registering cmd2 input with event loops, like asyncio, without occupying the main loop.
"""

import cmd2


class Cmd2EventBased(cmd2.Cmd):
    """Basic example of how to run cmd2 without it controlling the main loop."""

    def __init__(self) -> None:
        super().__init__()

    # ... your class code here ...


if __name__ == '__main__':
    app = Cmd2EventBased()
    app.preloop()

    # Do this within whatever event loop mechanism you wish to run a single command.
    # In this case, no prompt is generated, so you need to provide one and read the user's input.
    app.onecmd_plus_hooks("help history")

    app.postloop()



#!/usr/bin/env python
"""A simple example demonstrating the using paged output via the ppaged() method."""

import os

import cmd2


class PagedOutput(cmd2.Cmd):
    """Example cmd2 application which shows how to display output using a pager."""

    def __init__(self) -> None:
        super().__init__()

    def page_file(self, file_path: str, chop: bool = False) -> None:
        """Helper method to prevent having too much duplicated code."""
        filename = os.path.expanduser(file_path)
        try:
            with open(filename) as f:
                text = f.read()
            self.ppaged(text, chop=chop)
        except OSError as ex:
            self.pexcept(f'Error reading {filename!r}: {ex}')

    @cmd2.with_argument_list
    def do_page_wrap(self, args: list[str]) -> None:
        """Read in a text file and display its output in a pager, wrapping long lines if they don't fit.

        Usage: page_wrap <file_path>
        """
        if not args:
            self.perror('page_wrap requires a path to a file as an argument')
            return
        self.page_file(args[0], chop=False)

    complete_page_wrap = cmd2.Cmd.path_complete

    @cmd2.with_argument_list
    def do_page_truncate(self, args: list[str]) -> None:
        """Read in a text file and display its output in a pager, truncating long lines if they don't fit.

        Truncated lines can still be accessed by scrolling to the right using the arrow keys.

        Usage: page_chop <file_path>
        """
        if not args:
            self.perror('page_truncate requires a path to a file as an argument')
            return
        self.page_file(args[0], chop=True)

    complete_page_truncate = cmd2.Cmd.path_complete


if __name__ == '__main__':
    import sys

    app = PagedOutput()
    sys.exit(app.cmdloop())



"""Unit testing for cmd2/utils.py module."""

import os
import signal
import sys
import time
from unittest import (
    mock,
)

import pytest

import cmd2.utils as cu

HELLO_WORLD = 'Hello, world!'


def test_remove_duplicates_no_duplicates() -> None:
    no_dups = [5, 4, 3, 2, 1]
    assert cu.remove_duplicates(no_dups) == no_dups


def test_remove_duplicates_with_duplicates() -> None:
    duplicates = [1, 1, 2, 3, 9, 9, 7, 8]
    assert cu.remove_duplicates(duplicates) == [1, 2, 3, 9, 7, 8]


def test_alphabetical_sort() -> None:
    my_list = ['café', 'µ', 'A', 'micro', 'unity', 'cafeteria']
    assert cu.alphabetical_sort(my_list) == ['A', 'cafeteria', 'café', 'micro', 'unity', 'µ']
    my_list = ['a3', 'a22', 'A2', 'A11', 'a1']
    assert cu.alphabetical_sort(my_list) == ['a1', 'A11', 'A2', 'a22', 'a3']


def test_try_int_or_force_to_lower_case() -> None:
    str1 = '17'
    assert cu.try_int_or_force_to_lower_case(str1) == 17
    str1 = 'ABC'
    assert cu.try_int_or_force_to_lower_case(str1) == 'abc'
    str1 = 'X19'
    assert cu.try_int_or_force_to_lower_case(str1) == 'x19'
    str1 = ''
    assert cu.try_int_or_force_to_lower_case(str1) == ''


def test_natural_keys() -> None:
    my_list = ['café', 'µ', 'A', 'micro', 'unity', 'x1', 'X2', 'X11', 'X0', 'x22']
    my_list.sort(key=cu.natural_keys)
    assert my_list == ['A', 'café', 'micro', 'unity', 'X0', 'x1', 'X2', 'X11', 'x22', 'µ']
    my_list = ['a3', 'a22', 'A2', 'A11', 'a1']
    my_list.sort(key=cu.natural_keys)
    assert my_list == ['a1', 'A2', 'a3', 'A11', 'a22']


def test_natural_sort() -> None:
    my_list = ['café', 'µ', 'A', 'micro', 'unity', 'x1', 'X2', 'X11', 'X0', 'x22']
    assert cu.natural_sort(my_list) == ['A', 'café', 'micro', 'unity', 'X0', 'x1', 'X2', 'X11', 'x22', 'µ']
    my_list = ['a3', 'a22', 'A2', 'A11', 'a1']
    assert cu.natural_sort(my_list) == ['a1', 'A2', 'a3', 'A11', 'a22']


@pytest.fixture
def stdout_sim():
    return cu.StdSim(sys.stdout, echo=True)


def test_stdsim_write_str(stdout_sim) -> None:
    my_str = 'Hello World'
    stdout_sim.write(my_str)
    assert stdout_sim.getvalue() == my_str


def test_stdsim_write_bytes(stdout_sim) -> None:
    b_str = b'Hello World'
    with pytest.raises(TypeError):
        stdout_sim.write(b_str)


def test_stdsim_buffer_write_bytes(stdout_sim) -> None:
    b_str = b'Hello World'
    stdout_sim.buffer.write(b_str)
    assert stdout_sim.getvalue() == b_str.decode()
    assert stdout_sim.getbytes() == b_str


def test_stdsim_buffer_write_str(stdout_sim) -> None:
    my_str = 'Hello World'
    with pytest.raises(TypeError):
        stdout_sim.buffer.write(my_str)


def test_stdsim_read(stdout_sim) -> None:
    my_str = 'Hello World'
    stdout_sim.write(my_str)
    # getvalue() returns the value and leaves it unaffected internally
    assert stdout_sim.getvalue() == my_str
    # read() returns the value and then clears the internal buffer
    assert stdout_sim.read() == my_str
    assert stdout_sim.getvalue() == ''

    stdout_sim.write(my_str)

    assert stdout_sim.getvalue() == my_str
    assert stdout_sim.read(2) == my_str[:2]
    assert stdout_sim.getvalue() == my_str[2:]


def test_stdsim_read_bytes(stdout_sim) -> None:
    b_str = b'Hello World'
    stdout_sim.buffer.write(b_str)
    # getbytes() returns the value and leaves it unaffected internally
    assert stdout_sim.getbytes() == b_str
    # read_bytes() returns the value and then clears the internal buffer
    assert stdout_sim.readbytes() == b_str
    assert stdout_sim.getbytes() == b''


def test_stdsim_clear(stdout_sim) -> None:
    my_str = 'Hello World'
    stdout_sim.write(my_str)
    assert stdout_sim.getvalue() == my_str
    stdout_sim.clear()
    assert stdout_sim.getvalue() == ''


def test_stdsim_getattr_exist(stdout_sim) -> None:
    # Here the StdSim getattr is allowing us to access methods within StdSim
    my_str = 'Hello World'
    stdout_sim.write(my_str)
    val_func = stdout_sim.getvalue
    assert val_func() == my_str


def test_stdsim_getattr_noexist(stdout_sim) -> None:
    # Here the StdSim getattr is allowing us to access methods defined by the inner stream
    assert not stdout_sim.isatty()


def test_stdsim_pause_storage(stdout_sim) -> None:
    # Test pausing storage for string data
    my_str = 'Hello World'

    stdout_sim.pause_storage = False
    stdout_sim.write(my_str)
    assert stdout_sim.read() == my_str

    stdout_sim.pause_storage = True
    stdout_sim.write(my_str)
    assert stdout_sim.read() == ''

    # Test pausing storage for binary data
    b_str = b'Hello World'

    stdout_sim.pause_storage = False
    stdout_sim.buffer.write(b_str)
    assert stdout_sim.readbytes() == b_str

    stdout_sim.pause_storage = True
    stdout_sim.buffer.write(b_str)
    assert stdout_sim.getbytes() == b''


def test_stdsim_line_buffering(base_app) -> None:
    # This exercises the case of writing binary data that contains new lines/carriage returns to a StdSim
    # when line buffering is on. The output should immediately be flushed to the underlying stream.
    import os
    import tempfile

    with tempfile.NamedTemporaryFile(mode='wt') as file:
        file.line_buffering = True

        stdsim = cu.StdSim(file, echo=True)
        saved_size = os.path.getsize(file.name)

        bytes_to_write = b'hello\n'
        stdsim.buffer.write(bytes_to_write)
        assert os.path.getsize(file.name) == saved_size + len(bytes_to_write)
        saved_size = os.path.getsize(file.name)

        bytes_to_write = b'hello\r'
        stdsim.buffer.write(bytes_to_write)
        assert os.path.getsize(file.name) == saved_size + len(bytes_to_write)


@pytest.fixture
def pr_none():
    import subprocess

    # Start a long running process so we have time to run tests on it before it finishes
    # Put the new process into a separate group so its signal are isolated from ours
    kwargs = {}
    if sys.platform.startswith('win'):
        command = 'timeout -t 5 /nobreak'
        kwargs['creationflags'] = subprocess.CREATE_NEW_PROCESS_GROUP
    else:
        command = 'sleep 5'
        kwargs['start_new_session'] = True

    proc = subprocess.Popen(command, shell=True, **kwargs)
    return cu.ProcReader(proc, None, None)


def test_proc_reader_send_sigint(pr_none) -> None:
    assert pr_none._proc.poll() is None
    pr_none.send_sigint()
    pr_none.wait()
    ret_code = pr_none._proc.poll()

    # Make sure a SIGINT killed the process
    if sys.platform.startswith('win'):
        assert ret_code is not None
    else:
        assert ret_code == -signal.SIGINT


def test_proc_reader_terminate(pr_none) -> None:
    assert pr_none._proc.poll() is None
    pr_none.terminate()

    wait_start = time.monotonic()
    pr_none.wait()
    wait_finish = time.monotonic()

    # Make sure the process exited before sleep of 5 seconds finished
    # 3 seconds accounts for some delay but is long enough for the process to exit
    assert wait_finish - wait_start < 3

    ret_code = pr_none._proc.poll()
    if sys.platform.startswith('win'):
        assert ret_code is not None
    else:
        assert ret_code == -signal.SIGTERM


@pytest.fixture
def context_flag():
    return cu.ContextFlag()


def test_context_flag_bool(context_flag) -> None:
    assert not context_flag
    with context_flag:
        assert context_flag


def test_context_flag_exit_err(context_flag) -> None:
    with pytest.raises(ValueError, match="count has gone below 0"):
        context_flag.__exit__()


def test_to_bool_str_true() -> None:
    assert cu.to_bool('true')
    assert cu.to_bool('True')
    assert cu.to_bool('TRUE')
    assert cu.to_bool('tRuE')


def test_to_bool_str_false() -> None:
    assert not cu.to_bool('false')
    assert not cu.to_bool('False')
    assert not cu.to_bool('FALSE')
    assert not cu.to_bool('fAlSe')


def test_to_bool_str_invalid() -> None:
    with pytest.raises(ValueError):  # noqa: PT011
        cu.to_bool('other')


def test_to_bool_bool() -> None:
    assert cu.to_bool(True)
    assert not cu.to_bool(False)


def test_to_bool_int() -> None:
    assert cu.to_bool(1)
    assert cu.to_bool(-1)
    assert not cu.to_bool(0)


def test_to_bool_float() -> None:
    assert cu.to_bool(2.35)
    assert cu.to_bool(0.25)
    assert cu.to_bool(-3.1415)
    assert not cu.to_bool(0)


def test_find_editor_specified() -> None:
    expected_editor = os.path.join('fake_dir', 'editor')
    with mock.patch.dict(os.environ, {'EDITOR': expected_editor}):
        editor = cu.find_editor()
    assert editor == expected_editor


def test_find_editor_not_specified() -> None:
    # Use existing path env setting. Something in the editor list should be found.
    editor = cu.find_editor()
    assert editor

    # Overwrite path env setting with invalid path, clear all other env vars so no editor should be found.
    with mock.patch.dict(os.environ, {'PATH': 'fake_dir'}, clear=True):
        editor = cu.find_editor()
    assert editor is None


def test_similarity() -> None:
    suggested_command = cu.suggest_similar("comand", ["command", "UNRELATED", "NOT_SIMILAR"])
    assert suggested_command == "command"
    suggested_command = cu.suggest_similar("command", ["COMMAND", "acommands"])
    assert suggested_command == "COMMAND"


def test_similarity_without_good_canididates() -> None:
    suggested_command = cu.suggest_similar("comand", ["UNRELATED", "NOT_SIMILAR"])
    assert suggested_command is None
    suggested_command = cu.suggest_similar("comand", [])
    assert suggested_command is None


def test_similarity_overwrite_function() -> None:
    options = ["history", "test"]
    suggested_command = cu.suggest_similar("test", options)
    assert suggested_command == 'test'

    def custom_similarity_function(s1, s2) -> float:
        return 1.0 if 'history' in (s1, s2) else 0.0

    suggested_command = cu.suggest_similar("test", options, similarity_function_to_use=custom_similarity_function)
    assert suggested_command == 'history'

    suggested_command = cu.suggest_similar("history", options, similarity_function_to_use=custom_similarity_function)
    assert suggested_command == 'history'

    suggested_command = cu.suggest_similar("test", ["test"], similarity_function_to_use=custom_similarity_function)
    assert suggested_command is None


def test_get_types_invalid_input() -> None:
    x = 1
    with pytest.raises(ValueError, match="Argument passed to get_types should be a function or method"):
        cu.get_types(x)


def test_get_types_empty() -> None:
    def a(b):
        print(b)

    param_ann, ret_ann = cu.get_types(a)
    assert ret_ann is None
    assert param_ann == {}


def test_get_types_non_empty() -> None:
    def foo(x: int) -> str:
        return f"{x * x}"

    param_ann, ret_ann = cu.get_types(foo)
    assert ret_ann is str
    param_name, param_value = next(iter(param_ann.items()))
    assert param_name == 'x'
    assert param_value is int


def test_get_types_method() -> None:
    class Foo:
        def bar(self, x: bool) -> None:
            print(x)

    f = Foo()

    param_ann, ret_ann = cu.get_types(f.bar)
    assert ret_ann is None
    assert len(param_ann) == 1
    param_name, param_value = next(iter(param_ann.items()))
    assert param_name == 'x'
    assert param_value is bool



"""Unit/functional testing for readline tab completion functions in the cmd2.py module.

These are primarily tests related to readline completer functions which handle tab completion of cmd2/cmd commands,
file system paths, and shell commands.
"""

import enum
import os
import sys
from typing import NoReturn
from unittest import mock

import pytest

import cmd2
from cmd2 import utils

from .conftest import (
    complete_tester,
    normalize,
    run_cmd,
)


class SubcommandsExample(cmd2.Cmd):
    """Example cmd2 application where we a base command which has a couple subcommands
    and the "sport" subcommand has tab completion enabled.
    """

    sport_item_strs = ('Bat', 'Basket', 'Basketball', 'Football', 'Space Ball')

    # create the top-level parser for the base command
    base_parser = cmd2.Cmd2ArgumentParser()
    base_subparsers = base_parser.add_subparsers(title='subcommands', help='subcommand help')

    # create the parser for the "foo" subcommand
    parser_foo = base_subparsers.add_parser('foo', help='foo help')
    parser_foo.add_argument('-x', type=int, default=1, help='integer')
    parser_foo.add_argument('y', type=float, help='float')
    parser_foo.add_argument('input_file', type=str, help='Input File')

    # create the parser for the "bar" subcommand
    parser_bar = base_subparsers.add_parser('bar', help='bar help')

    bar_subparsers = parser_bar.add_subparsers(title='layer3', help='help for 3rd layer of commands')
    parser_bar.add_argument('z', help='string')

    bar_subparsers.add_parser('apple', help='apple help')
    bar_subparsers.add_parser('artichoke', help='artichoke help')
    bar_subparsers.add_parser('cranberries', help='cranberries help')

    # create the parser for the "sport" subcommand
    parser_sport = base_subparsers.add_parser('sport', help='sport help')
    sport_arg = parser_sport.add_argument('sport', help='Enter name of a sport', choices=sport_item_strs)

    # create the top-level parser for the alternate command
    # The alternate command doesn't provide its own help flag
    base2_parser = cmd2.Cmd2ArgumentParser(add_help=False)
    base2_subparsers = base2_parser.add_subparsers(title='subcommands', help='subcommand help')

    # create the parser for the "foo" subcommand
    parser_foo2 = base2_subparsers.add_parser('foo', help='foo help')
    parser_foo2.add_argument('-x', type=int, default=1, help='integer')
    parser_foo2.add_argument('y', type=float, help='float')
    parser_foo2.add_argument('input_file', type=str, help='Input File')

    # create the parser for the "bar" subcommand
    parser_bar2 = base2_subparsers.add_parser('bar', help='bar help')

    bar2_subparsers = parser_bar2.add_subparsers(title='layer3', help='help for 3rd layer of commands')
    parser_bar2.add_argument('z', help='string')

    bar2_subparsers.add_parser('apple', help='apple help')
    bar2_subparsers.add_parser('artichoke', help='artichoke help')
    bar2_subparsers.add_parser('cranberries', help='cranberries help')

    # create the parser for the "sport" subcommand
    parser_sport2 = base2_subparsers.add_parser('sport', help='sport help')
    sport2_arg = parser_sport2.add_argument('sport', help='Enter name of a sport', choices=sport_item_strs)

    def __init__(self) -> None:
        super().__init__()

    # subcommand functions for the base command
    def base_foo(self, args) -> None:
        """Foo subcommand of base command."""
        self.poutput(args.x * args.y)

    def base_bar(self, args) -> None:
        """Bar subcommand of base command."""
        self.poutput(f'(({args.z}))')

    def base_sport(self, args) -> None:
        """Sport subcommand of base command."""
        self.poutput(f'Sport is {args.sport}')

    # Set handler functions for the subcommands
    parser_foo.set_defaults(func=base_foo)
    parser_bar.set_defaults(func=base_bar)
    parser_sport.set_defaults(func=base_sport)

    @cmd2.with_argparser(base_parser)
    def do_base(self, args) -> None:
        """Base command help."""
        func = getattr(args, 'func', None)
        if func is not None:
            # Call whatever subcommand function was selected
            func(self, args)
        else:
            # No subcommand was provided, so call help
            self.do_help('base')

    @cmd2.with_argparser(base2_parser)
    def do_alternate(self, args) -> None:
        """Alternate command help."""
        func = getattr(args, 'func', None)
        if func is not None:
            # Call whatever subcommand function was selected
            func(self, args)
        else:
            # No subcommand was provided, so call help
            self.do_help('alternate')


# List of strings used with completion functions
food_item_strs = ['Pizza', 'Ham', 'Ham Sandwich', 'Potato', 'Cheese "Pizza"']
sport_item_strs = ['Bat', 'Basket', 'Basketball', 'Football', 'Space Ball']
delimited_strs = [
    '/home/user/file.txt',
    '/home/user/file space.txt',
    '/home/user/prog.c',
    '/home/other user/maps',
    '/home/other user/tests',
]

# Dictionary used with flag based completion functions
flag_dict = {
    # Tab complete food items after -f and --food flag in command line
    '-f': food_item_strs,
    '--food': food_item_strs,
    # Tab complete sport items after -s and --sport flag in command line
    '-s': sport_item_strs,
    '--sport': sport_item_strs,
}

# Dictionary used with index based completion functions
index_dict = {
    1: food_item_strs,  # Tab complete food items at index 1 in command line
    2: sport_item_strs,  # Tab complete sport items at index 2 in command line
}


class CompletionsExample(cmd2.Cmd):
    """Example cmd2 application used to exercise tab completion tests"""

    def __init__(self) -> None:
        cmd2.Cmd.__init__(self, multiline_commands=['test_multiline'])
        self.foo = 'bar'
        self.add_settable(
            utils.Settable(
                'foo',
                str,
                description="a settable param",
                settable_object=self,
                completer=CompletionsExample.complete_foo_val,
            )
        )

    def do_test_basic(self, args) -> None:
        pass

    def complete_test_basic(self, text, line, begidx, endidx):
        return self.basic_complete(text, line, begidx, endidx, food_item_strs)

    def do_test_delimited(self, args) -> None:
        pass

    def complete_test_delimited(self, text, line, begidx, endidx):
        return self.delimiter_complete(text, line, begidx, endidx, delimited_strs, '/')

    def do_test_sort_key(self, args) -> None:
        pass

    def complete_test_sort_key(self, text, line, begidx, endidx):
        num_strs = ['2', '11', '1']
        return self.basic_complete(text, line, begidx, endidx, num_strs)

    def do_test_raise_exception(self, args) -> None:
        pass

    def complete_test_raise_exception(self, text, line, begidx, endidx) -> NoReturn:
        raise IndexError("You are out of bounds!!")

    def do_test_multiline(self, args) -> None:
        pass

    def complete_test_multiline(self, text, line, begidx, endidx):
        return self.basic_complete(text, line, begidx, endidx, sport_item_strs)

    def do_test_no_completer(self, args) -> None:
        """Completing this should result in completedefault() being called"""

    def complete_foo_val(self, text, line, begidx, endidx, arg_tokens):
        """Supports unit testing cmd2.Cmd2.complete_set_val to confirm it passes all tokens in the set command"""
        if 'param' in arg_tokens:
            return ["SUCCESS"]
        return ["FAIL"]

    def completedefault(self, *ignored):
        """Method called to complete an input line when no command-specific
        complete_*() method is available.

        """
        return ['default']


@pytest.fixture
def cmd2_app():
    return CompletionsExample()


def test_complete_command_single(cmd2_app) -> None:
    text = 'he'
    line = text
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert cmd2_app.completion_matches == ['help ']


def test_complete_empty_arg(cmd2_app) -> None:
    text = ''
    line = f'help {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    expected = sorted(cmd2_app.get_visible_commands(), key=cmd2_app.default_sort_key)
    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)

    assert first_match is not None
    assert cmd2_app.completion_matches == expected


def test_complete_bogus_command(cmd2_app) -> None:
    text = ''
    line = f'fizbuzz {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    expected = ['default ']
    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert cmd2_app.completion_matches == expected


def test_complete_exception(cmd2_app, capsys) -> None:
    text = ''
    line = f'test_raise_exception {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    _out, err = capsys.readouterr()

    assert first_match is None
    assert "IndexError" in err


def test_complete_macro(base_app, request) -> None:
    # Create the macro
    out, _err = run_cmd(base_app, 'macro create fake run_pyscript {1}')
    assert out == normalize("Macro 'fake' created")

    # Macros do path completion
    test_dir = os.path.dirname(request.module.__file__)

    text = os.path.join(test_dir, 's')
    line = f'fake {text}'

    endidx = len(line)
    begidx = endidx - len(text)

    expected = [text + 'cript.py', text + 'cript.txt', text + 'cripts' + os.path.sep]
    first_match = complete_tester(text, line, begidx, endidx, base_app)
    assert first_match is not None
    assert base_app.completion_matches == expected


def test_default_sort_key(cmd2_app) -> None:
    text = ''
    line = f'test_sort_key {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    # First do alphabetical sorting
    cmd2_app.default_sort_key = cmd2.Cmd.ALPHABETICAL_SORT_KEY
    expected = ['1', '11', '2']
    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert cmd2_app.completion_matches == expected

    # Now switch to natural sorting
    cmd2_app.default_sort_key = cmd2.Cmd.NATURAL_SORT_KEY
    expected = ['1', '2', '11']
    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert cmd2_app.completion_matches == expected


def test_cmd2_command_completion_multiple(cmd2_app) -> None:
    text = 'h'
    line = text
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert cmd2_app.completion_matches == ['help', 'history']


def test_cmd2_command_completion_nomatch(cmd2_app) -> None:
    text = 'fakecommand'
    line = text
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is None
    assert cmd2_app.completion_matches == []


def test_cmd2_help_completion_single(cmd2_app) -> None:
    text = 'he'
    line = f'help {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)

    # It is at end of line, so extra space is present
    assert first_match is not None
    assert cmd2_app.completion_matches == ['help ']


def test_cmd2_help_completion_multiple(cmd2_app) -> None:
    text = 'h'
    line = f'help {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert cmd2_app.completion_matches == ['help', 'history']


def test_cmd2_help_completion_nomatch(cmd2_app) -> None:
    text = 'fakecommand'
    line = f'help {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is None


def test_set_allow_style_completion(cmd2_app) -> None:
    """Confirm that completing allow_style presents AllowStyle strings"""
    text = ''
    line = 'set allow_style'
    endidx = len(line)
    begidx = endidx - len(text)

    expected = [val.name.lower() for val in cmd2.rich_utils.AllowStyle]

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match
    assert cmd2_app.completion_matches == sorted(expected, key=cmd2_app.default_sort_key)


def test_set_bool_completion(cmd2_app) -> None:
    """Confirm that completing a boolean Settable presents true and false strings"""
    text = ''
    line = 'set debug'
    endidx = len(line)
    begidx = endidx - len(text)

    expected = ['false', 'true']

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match
    assert cmd2_app.completion_matches == sorted(expected, key=cmd2_app.default_sort_key)


def test_shell_command_completion_shortcut(cmd2_app) -> None:
    # Made sure ! runs a shell command and all matches start with ! since there
    # isn't a space between ! and the shell command. Display matches won't
    # begin with the !.
    if sys.platform == "win32":
        text = '!calc'
        expected = ['!calc.exe ']
        expected_display = ['calc.exe']
    else:
        text = '!egr'
        expected = ['!egrep ']
        expected_display = ['egrep']

    line = text
    endidx = len(line)
    begidx = 0

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert cmd2_app.completion_matches == expected
    assert cmd2_app.display_matches == expected_display


def test_shell_command_completion_doesnt_match_wildcards(cmd2_app) -> None:
    if sys.platform == "win32":
        text = 'c*'
    else:
        text = 'e*'

    line = f'shell {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is None


def test_shell_command_completion_multiple(cmd2_app) -> None:
    if sys.platform == "win32":
        text = 'c'
        expected = 'calc.exe'
    else:
        text = 'l'
        expected = 'ls'

    line = f'shell {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert expected in cmd2_app.completion_matches


def test_shell_command_completion_nomatch(cmd2_app) -> None:
    text = 'zzzz'
    line = f'shell {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is None


def test_shell_command_completion_doesnt_complete_when_just_shell(cmd2_app) -> None:
    text = ''
    line = f'shell {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is None


def test_shell_command_completion_does_path_completion_when_after_command(cmd2_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)

    text = os.path.join(test_dir, 'conftest')
    line = f'shell cat {text}'

    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert cmd2_app.completion_matches == [text + '.py ']


def test_shell_command_complete_in_path(cmd2_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)

    text = os.path.join(test_dir, 's')
    line = f'shell {text}'

    endidx = len(line)
    begidx = endidx - len(text)

    # Since this will look for directories and executables in the given path,
    # we expect to see the scripts dir among the results
    expected = os.path.join(test_dir, 'scripts' + os.path.sep)
    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert expected in cmd2_app.completion_matches


def test_path_completion_single_end(cmd2_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)

    text = os.path.join(test_dir, 'conftest')
    line = f'shell cat {text}'

    endidx = len(line)
    begidx = endidx - len(text)

    assert cmd2_app.path_complete(text, line, begidx, endidx) == [text + '.py']


def test_path_completion_multiple(cmd2_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)

    text = os.path.join(test_dir, 's')
    line = f'shell cat {text}'

    endidx = len(line)
    begidx = endidx - len(text)

    matches = cmd2_app.path_complete(text, line, begidx, endidx)
    expected = [text + 'cript.py', text + 'cript.txt', text + 'cripts' + os.path.sep]
    assert matches == expected


def test_path_completion_nomatch(cmd2_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)

    text = os.path.join(test_dir, 'fakepath')
    line = f'shell cat {text}'

    endidx = len(line)
    begidx = endidx - len(text)

    assert cmd2_app.path_complete(text, line, begidx, endidx) == []


def test_default_to_shell_completion(cmd2_app, request) -> None:
    cmd2_app.default_to_shell = True
    test_dir = os.path.dirname(request.module.__file__)

    text = os.path.join(test_dir, 'conftest')

    if sys.platform == "win32":
        command = 'calc.exe'
    else:
        command = 'egrep'

    # Make sure the command is on the testing system
    assert command in utils.get_exes_in_path(command)
    line = f'{command} {text}'

    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert cmd2_app.completion_matches == [text + '.py ']


def test_path_completion_no_text(cmd2_app) -> None:
    # Run path complete with no search text which should show what's in cwd
    text = ''
    line = f'shell ls {text}'
    endidx = len(line)
    begidx = endidx - len(text)
    completions_no_text = cmd2_app.path_complete(text, line, begidx, endidx)

    # Run path complete with path set to the CWD
    text = os.getcwd() + os.path.sep
    line = f'shell ls {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    # We have to strip off the path from the beginning since the matches are entire paths
    completions_cwd = [match.replace(text, '', 1) for match in cmd2_app.path_complete(text, line, begidx, endidx)]

    # Verify that the first test gave results for entries in the cwd
    assert completions_no_text == completions_cwd
    assert completions_cwd


def test_path_completion_no_path(cmd2_app) -> None:
    # Run path complete with search text that isn't preceded by a path. This should use CWD as the path.
    text = 'p'
    line = f'shell ls {text}'
    endidx = len(line)
    begidx = endidx - len(text)
    completions_no_text = cmd2_app.path_complete(text, line, begidx, endidx)

    # Run path complete with path set to the CWD
    text = os.getcwd() + os.path.sep + text
    line = f'shell ls {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    # We have to strip off the path from the beginning since the matches are entire paths (Leave the 's')
    completions_cwd = [match.replace(text[:-1], '', 1) for match in cmd2_app.path_complete(text, line, begidx, endidx)]

    # Verify that the first test gave results for entries in the cwd
    assert completions_no_text == completions_cwd
    assert completions_cwd


@pytest.mark.skipif(sys.platform == 'win32', reason="this only applies on systems where the root directory is a slash")
def test_path_completion_cwd_is_root_dir(cmd2_app) -> None:
    # Change our CWD to root dir
    cwd = os.getcwd()
    os.chdir(os.path.sep)

    text = ''
    line = f'shell ls {text}'
    endidx = len(line)
    begidx = endidx - len(text)
    completions = cmd2_app.path_complete(text, line, begidx, endidx)

    # No match should start with a slash
    assert not any(match.startswith(os.path.sep) for match in completions)

    # Restore CWD
    os.chdir(cwd)


def test_path_completion_doesnt_match_wildcards(cmd2_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)

    text = os.path.join(test_dir, 'c*')
    line = f'shell cat {text}'

    endidx = len(line)
    begidx = endidx - len(text)

    # Currently path completion doesn't accept wildcards, so will always return empty results
    assert cmd2_app.path_complete(text, line, begidx, endidx) == []


def test_path_completion_complete_user(cmd2_app) -> None:
    import getpass

    user = getpass.getuser()

    text = f'~{user}'
    line = f'shell fake {text}'
    endidx = len(line)
    begidx = endidx - len(text)
    completions = cmd2_app.path_complete(text, line, begidx, endidx)

    expected = text + os.path.sep
    assert expected in completions


def test_path_completion_user_path_expansion(cmd2_app) -> None:
    # Run path with a tilde and a slash
    if sys.platform.startswith('win'):
        cmd = 'dir'
    else:
        cmd = 'ls'

    # Use a ~ which will be expanded into the user's home directory
    text = f'~{os.path.sep}'
    line = f'shell {cmd} {text}'
    endidx = len(line)
    begidx = endidx - len(text)
    completions_tilde_slash = [match.replace(text, '', 1) for match in cmd2_app.path_complete(text, line, begidx, endidx)]

    # Run path complete on the user's home directory
    text = os.path.expanduser('~') + os.path.sep
    line = f'shell {cmd} {text}'
    endidx = len(line)
    begidx = endidx - len(text)
    completions_home = [match.replace(text, '', 1) for match in cmd2_app.path_complete(text, line, begidx, endidx)]

    assert completions_tilde_slash == completions_home


def test_path_completion_directories_only(cmd2_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)

    text = os.path.join(test_dir, 's')
    line = f'shell cat {text}'

    endidx = len(line)
    begidx = endidx - len(text)

    expected = [text + 'cripts' + os.path.sep]

    assert cmd2_app.path_complete(text, line, begidx, endidx, path_filter=os.path.isdir) == expected


def test_basic_completion_single(cmd2_app) -> None:
    text = 'Pi'
    line = f'list_food -f {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    assert cmd2_app.basic_complete(text, line, begidx, endidx, food_item_strs) == ['Pizza']


def test_basic_completion_multiple(cmd2_app) -> None:
    text = ''
    line = f'list_food -f {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    matches = sorted(cmd2_app.basic_complete(text, line, begidx, endidx, food_item_strs))
    assert matches == sorted(food_item_strs)


def test_basic_completion_nomatch(cmd2_app) -> None:
    text = 'q'
    line = f'list_food -f {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    assert cmd2_app.basic_complete(text, line, begidx, endidx, food_item_strs) == []


def test_delimiter_completion_partial(cmd2_app) -> None:
    """Test that a delimiter is added when an item has not been fully completed"""
    text = '/home/'
    line = f'command {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    matches = cmd2_app.delimiter_complete(text, line, begidx, endidx, delimited_strs, '/')

    # All matches end with the delimiter
    matches.sort(key=cmd2_app.default_sort_key)
    expected_matches = sorted(["/home/other user/", "/home/user/"], key=cmd2_app.default_sort_key)

    cmd2_app.display_matches.sort(key=cmd2_app.default_sort_key)
    expected_display = sorted(["other user/", "user/"], key=cmd2_app.default_sort_key)

    assert matches == expected_matches
    assert cmd2_app.display_matches == expected_display


def test_delimiter_completion_full(cmd2_app) -> None:
    """Test that no delimiter is added when an item has been fully completed"""
    text = '/home/other user/'
    line = f'command {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    matches = cmd2_app.delimiter_complete(text, line, begidx, endidx, delimited_strs, '/')

    # No matches end with the delimiter
    matches.sort(key=cmd2_app.default_sort_key)
    expected_matches = sorted(["/home/other user/maps", "/home/other user/tests"], key=cmd2_app.default_sort_key)

    cmd2_app.display_matches.sort(key=cmd2_app.default_sort_key)
    expected_display = sorted(["maps", "tests"], key=cmd2_app.default_sort_key)

    assert matches == expected_matches
    assert cmd2_app.display_matches == expected_display


def test_delimiter_completion_nomatch(cmd2_app) -> None:
    text = '/nothing_to_see'
    line = f'command {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    assert cmd2_app.delimiter_complete(text, line, begidx, endidx, delimited_strs, '/') == []


def test_flag_based_completion_single(cmd2_app) -> None:
    text = 'Pi'
    line = f'list_food -f {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    assert cmd2_app.flag_based_complete(text, line, begidx, endidx, flag_dict) == ['Pizza']


def test_flag_based_completion_multiple(cmd2_app) -> None:
    text = ''
    line = f'list_food -f {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    matches = sorted(cmd2_app.flag_based_complete(text, line, begidx, endidx, flag_dict))
    assert matches == sorted(food_item_strs)


def test_flag_based_completion_nomatch(cmd2_app) -> None:
    text = 'q'
    line = f'list_food -f {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    assert cmd2_app.flag_based_complete(text, line, begidx, endidx, flag_dict) == []


def test_flag_based_default_completer(cmd2_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)

    text = os.path.join(test_dir, 'c')
    line = f'list_food {text}'

    endidx = len(line)
    begidx = endidx - len(text)

    assert cmd2_app.flag_based_complete(text, line, begidx, endidx, flag_dict, all_else=cmd2_app.path_complete) == [
        text + 'onftest.py'
    ]


def test_flag_based_callable_completer(cmd2_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)

    text = os.path.join(test_dir, 'c')
    line = f'list_food -o {text}'

    endidx = len(line)
    begidx = endidx - len(text)

    flag_dict['-o'] = cmd2_app.path_complete
    assert cmd2_app.flag_based_complete(text, line, begidx, endidx, flag_dict) == [text + 'onftest.py']


def test_index_based_completion_single(cmd2_app) -> None:
    text = 'Foo'
    line = f'command Pizza {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    assert cmd2_app.index_based_complete(text, line, begidx, endidx, index_dict) == ['Football']


def test_index_based_completion_multiple(cmd2_app) -> None:
    text = ''
    line = f'command Pizza {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    matches = sorted(cmd2_app.index_based_complete(text, line, begidx, endidx, index_dict))
    assert matches == sorted(sport_item_strs)


def test_index_based_completion_nomatch(cmd2_app) -> None:
    text = 'q'
    line = f'command {text}'
    endidx = len(line)
    begidx = endidx - len(text)
    assert cmd2_app.index_based_complete(text, line, begidx, endidx, index_dict) == []


def test_index_based_default_completer(cmd2_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)

    text = os.path.join(test_dir, 'c')
    line = f'command Pizza Bat Computer {text}'

    endidx = len(line)
    begidx = endidx - len(text)

    assert cmd2_app.index_based_complete(text, line, begidx, endidx, index_dict, all_else=cmd2_app.path_complete) == [
        text + 'onftest.py'
    ]


def test_index_based_callable_completer(cmd2_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)

    text = os.path.join(test_dir, 'c')
    line = f'command Pizza Bat {text}'

    endidx = len(line)
    begidx = endidx - len(text)

    index_dict[3] = cmd2_app.path_complete
    assert cmd2_app.index_based_complete(text, line, begidx, endidx, index_dict) == [text + 'onftest.py']


def test_tokens_for_completion_quoted(cmd2_app) -> None:
    text = 'Pi'
    line = f'list_food "{text}"'
    endidx = len(line)
    begidx = endidx

    expected_tokens = ['list_food', 'Pi', '']
    expected_raw_tokens = ['list_food', '"Pi"', '']

    tokens, raw_tokens = cmd2_app.tokens_for_completion(line, begidx, endidx)
    assert expected_tokens == tokens
    assert expected_raw_tokens == raw_tokens


def test_tokens_for_completion_unclosed_quote(cmd2_app) -> None:
    text = 'Pi'
    line = f'list_food "{text}'
    endidx = len(line)
    begidx = endidx - len(text)

    expected_tokens = ['list_food', 'Pi']
    expected_raw_tokens = ['list_food', '"Pi']

    tokens, raw_tokens = cmd2_app.tokens_for_completion(line, begidx, endidx)
    assert expected_tokens == tokens
    assert expected_raw_tokens == raw_tokens


def test_tokens_for_completion_punctuation(cmd2_app) -> None:
    """Test that redirectors and terminators are word delimiters"""
    text = 'file'
    line = f'command | < ;>>{text}'
    endidx = len(line)
    begidx = endidx - len(text)

    expected_tokens = ['command', '|', '<', ';', '>>', 'file']
    expected_raw_tokens = ['command', '|', '<', ';', '>>', 'file']

    tokens, raw_tokens = cmd2_app.tokens_for_completion(line, begidx, endidx)
    assert expected_tokens == tokens
    assert expected_raw_tokens == raw_tokens


def test_tokens_for_completion_quoted_punctuation(cmd2_app) -> None:
    """Test that quoted punctuation characters are not word delimiters"""
    text = '>file'
    line = f'command "{text}'
    endidx = len(line)
    begidx = endidx - len(text)

    expected_tokens = ['command', '>file']
    expected_raw_tokens = ['command', '">file']

    tokens, raw_tokens = cmd2_app.tokens_for_completion(line, begidx, endidx)
    assert expected_tokens == tokens
    assert expected_raw_tokens == raw_tokens


def test_add_opening_quote_basic_no_text(cmd2_app) -> None:
    text = ''
    line = f'test_basic {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    # The whole list will be returned with no opening quotes added
    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert cmd2_app.completion_matches == sorted(food_item_strs, key=cmd2_app.default_sort_key)


def test_add_opening_quote_basic_nothing_added(cmd2_app) -> None:
    text = 'P'
    line = f'test_basic {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert cmd2_app.completion_matches == ['Pizza', 'Potato']


def test_add_opening_quote_basic_quote_added(cmd2_app) -> None:
    text = 'Ha'
    line = f'test_basic {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    expected = sorted(['"Ham', '"Ham Sandwich'], key=cmd2_app.default_sort_key)
    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert cmd2_app.completion_matches == expected


def test_add_opening_quote_basic_single_quote_added(cmd2_app) -> None:
    text = 'Ch'
    line = f'test_basic {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    expected = ["'Cheese \"Pizza\"' "]
    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert cmd2_app.completion_matches == expected


def test_add_opening_quote_basic_text_is_common_prefix(cmd2_app) -> None:
    # This tests when the text entered is the same as the common prefix of the matches
    text = 'Ham'
    line = f'test_basic {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    expected = sorted(['"Ham', '"Ham Sandwich'], key=cmd2_app.default_sort_key)
    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert cmd2_app.completion_matches == expected


def test_add_opening_quote_delimited_no_text(cmd2_app) -> None:
    text = ''
    line = f'test_delimited {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    # Matches returned with no opening quote
    expected_matches = sorted(["/home/other user/", "/home/user/"], key=cmd2_app.default_sort_key)
    expected_display = sorted(["other user/", "user/"], key=cmd2_app.default_sort_key)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert cmd2_app.completion_matches == expected_matches
    assert cmd2_app.display_matches == expected_display


def test_add_opening_quote_delimited_nothing_added(cmd2_app) -> None:
    text = '/home/'
    line = f'test_delimited {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    expected_matches = sorted(['/home/other user/', '/home/user/'], key=cmd2_app.default_sort_key)
    expected_display = sorted(['other user/', 'user/'], key=cmd2_app.default_sort_key)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert cmd2_app.completion_matches == expected_matches
    assert cmd2_app.display_matches == expected_display


def test_add_opening_quote_delimited_quote_added(cmd2_app) -> None:
    text = '/home/user/fi'
    line = f'test_delimited {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    expected_common_prefix = '"/home/user/file'
    expected_display = sorted(['file.txt', 'file space.txt'], key=cmd2_app.default_sort_key)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert os.path.commonprefix(cmd2_app.completion_matches) == expected_common_prefix
    assert cmd2_app.display_matches == expected_display


def test_add_opening_quote_delimited_text_is_common_prefix(cmd2_app) -> None:
    # This tests when the text entered is the same as the common prefix of the matches
    text = '/home/user/file'
    line = f'test_delimited {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    expected_common_prefix = '"/home/user/file'
    expected_display = sorted(['file.txt', 'file space.txt'], key=cmd2_app.default_sort_key)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert os.path.commonprefix(cmd2_app.completion_matches) == expected_common_prefix
    assert cmd2_app.display_matches == expected_display


def test_add_opening_quote_delimited_space_in_prefix(cmd2_app) -> None:
    # This tests when a space appears before the part of the string that is the display match
    text = '/home/oth'
    line = f'test_delimited {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    expected_common_prefix = '"/home/other user/'
    expected_display = ['maps', 'tests']

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert os.path.commonprefix(cmd2_app.completion_matches) == expected_common_prefix
    assert cmd2_app.display_matches == expected_display


def test_no_completer(cmd2_app) -> None:
    text = ''
    line = f'test_no_completer {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    expected = ['default ']
    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is not None
    assert cmd2_app.completion_matches == expected


def test_wordbreak_in_command(cmd2_app) -> None:
    text = ''
    line = f'"{text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is None
    assert not cmd2_app.completion_matches


def test_complete_multiline_on_single_line(cmd2_app) -> None:
    text = ''
    line = f'test_multiline {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    expected = sorted(sport_item_strs, key=cmd2_app.default_sort_key)
    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)

    assert first_match is not None
    assert cmd2_app.completion_matches == expected


def test_complete_multiline_on_multiple_lines(cmd2_app) -> None:
    # Set the same variables _complete_statement() sets when a user is entering data at a continuation prompt
    cmd2_app._at_continuation_prompt = True
    cmd2_app._multiline_in_progress = "test_multiline\n"

    text = 'Ba'
    line = f'{text}'
    endidx = len(line)
    begidx = endidx - len(text)

    expected = sorted(['Bat', 'Basket', 'Basketball'], key=cmd2_app.default_sort_key)
    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)

    assert first_match is not None
    assert cmd2_app.completion_matches == expected


# Used by redirect_complete tests
class RedirCompType(enum.Enum):
    SHELL_CMD = (1,)
    PATH = (2,)
    DEFAULT = (3,)
    NONE = 4


@pytest.mark.parametrize(
    ('line', 'comp_type'),
    [
        ('fake', RedirCompType.DEFAULT),
        ('fake arg', RedirCompType.DEFAULT),
        ('fake |', RedirCompType.SHELL_CMD),
        ('fake | grep', RedirCompType.PATH),
        ('fake | grep arg', RedirCompType.PATH),
        ('fake | grep >', RedirCompType.PATH),
        ('fake | grep > >', RedirCompType.NONE),
        ('fake | grep > file', RedirCompType.NONE),
        ('fake | grep > file >', RedirCompType.NONE),
        ('fake | grep > file |', RedirCompType.SHELL_CMD),
        ('fake | grep > file | grep', RedirCompType.PATH),
        ('fake | |', RedirCompType.NONE),
        ('fake | >', RedirCompType.NONE),
        ('fake >', RedirCompType.PATH),
        ('fake >>', RedirCompType.PATH),
        ('fake > >', RedirCompType.NONE),
        ('fake > |', RedirCompType.SHELL_CMD),
        ('fake >> file |', RedirCompType.SHELL_CMD),
        ('fake >> file | grep', RedirCompType.PATH),
        ('fake > file', RedirCompType.NONE),
        ('fake > file >', RedirCompType.NONE),
        ('fake > file >>', RedirCompType.NONE),
    ],
)
def test_redirect_complete(cmd2_app, monkeypatch, line, comp_type) -> None:
    # Test both cases of allow_redirection
    cmd2_app.allow_redirection = True
    for _ in range(2):
        shell_cmd_complete_mock = mock.MagicMock(name='shell_cmd_complete')
        monkeypatch.setattr("cmd2.Cmd.shell_cmd_complete", shell_cmd_complete_mock)

        path_complete_mock = mock.MagicMock(name='path_complete')
        monkeypatch.setattr("cmd2.Cmd.path_complete", path_complete_mock)

        default_complete_mock = mock.MagicMock(name='fake_completer')

        text = ''
        line = f'{line} {text}'
        endidx = len(line)
        begidx = endidx - len(text)

        cmd2_app._redirect_complete(text, line, begidx, endidx, default_complete_mock)

        if comp_type == RedirCompType.SHELL_CMD:
            shell_cmd_complete_mock.assert_called_once()
        elif comp_type == RedirCompType.PATH:
            path_complete_mock.assert_called_once()
        elif comp_type == RedirCompType.DEFAULT:
            default_complete_mock.assert_called_once()
        else:
            shell_cmd_complete_mock.assert_not_called()
            path_complete_mock.assert_not_called()
            default_complete_mock.assert_not_called()

        # Do the next test with allow_redirection as False
        cmd2_app.allow_redirection = False
        if comp_type != RedirCompType.DEFAULT:
            comp_type = RedirCompType.NONE


def test_complete_set_value(cmd2_app) -> None:
    text = ''
    line = f'set foo {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match == "SUCCESS "
    assert cmd2_app.completion_hint == "Hint:\n  value  a settable param\n"


def test_complete_set_value_invalid_settable(cmd2_app, capsys) -> None:
    text = ''
    line = f'set fake {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, cmd2_app)
    assert first_match is None

    out, _err = capsys.readouterr()
    assert "fake is not a settable parameter" in out


@pytest.fixture
def sc_app():
    c = SubcommandsExample()
    c.stdout = utils.StdSim(c.stdout)
    return c


def test_cmd2_subcommand_completion_single_end(sc_app) -> None:
    text = 'f'
    line = f'base {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, sc_app)

    # It is at end of line, so extra space is present
    assert first_match is not None
    assert sc_app.completion_matches == ['foo ']


def test_cmd2_subcommand_completion_multiple(sc_app) -> None:
    text = ''
    line = f'base {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, sc_app)
    assert first_match is not None
    assert sc_app.completion_matches == ['bar', 'foo', 'sport']


def test_cmd2_subcommand_completion_nomatch(sc_app) -> None:
    text = 'z'
    line = f'base {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, sc_app)
    assert first_match is None


def test_help_subcommand_completion_single(sc_app) -> None:
    text = 'base'
    line = f'help {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, sc_app)

    # It is at end of line, so extra space is present
    assert first_match is not None
    assert sc_app.completion_matches == ['base ']


def test_help_subcommand_completion_multiple(sc_app) -> None:
    text = ''
    line = f'help base {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, sc_app)
    assert first_match is not None
    assert sc_app.completion_matches == ['bar', 'foo', 'sport']


def test_help_subcommand_completion_nomatch(sc_app) -> None:
    text = 'z'
    line = f'help base {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, sc_app)
    assert first_match is None


def test_subcommand_tab_completion(sc_app) -> None:
    # This makes sure the correct completer for the sport subcommand is called
    text = 'Foot'
    line = f'base sport {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, sc_app)

    # It is at end of line, so extra space is present
    assert first_match is not None
    assert sc_app.completion_matches == ['Football ']


def test_subcommand_tab_completion_with_no_completer(sc_app) -> None:
    # This tests what happens when a subcommand has no completer
    # In this case, the foo subcommand has no completer defined
    text = 'Foot'
    line = f'base foo {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, sc_app)
    assert first_match is None


def test_subcommand_tab_completion_space_in_text(sc_app) -> None:
    text = 'B'
    line = f'base sport "Space {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, sc_app)

    assert first_match is not None
    assert sc_app.completion_matches == ['Ball" ']
    assert sc_app.display_matches == ['Space Ball']


####################################################


class SubcommandsWithUnknownExample(cmd2.Cmd):
    """Example cmd2 application where we a base command which has a couple subcommands
    and the "sport" subcommand has tab completion enabled.
    """

    def __init__(self) -> None:
        cmd2.Cmd.__init__(self)

    # subcommand functions for the base command
    def base_foo(self, args) -> None:
        """Foo subcommand of base command"""
        self.poutput(args.x * args.y)

    def base_bar(self, args) -> None:
        """Bar subcommand of base command"""
        self.poutput(f'(({args.z}))')

    def base_sport(self, args) -> None:
        """Sport subcommand of base command"""
        self.poutput(f'Sport is {args.sport}')

    # create the top-level parser for the base command
    base_parser = cmd2.Cmd2ArgumentParser()
    base_subparsers = base_parser.add_subparsers(title='subcommands', help='subcommand help')

    # create the parser for the "foo" subcommand
    parser_foo = base_subparsers.add_parser('foo', help='foo help')
    parser_foo.add_argument('-x', type=int, default=1, help='integer')
    parser_foo.add_argument('y', type=float, help='float')
    parser_foo.set_defaults(func=base_foo)

    # create the parser for the "bar" subcommand
    parser_bar = base_subparsers.add_parser('bar', help='bar help')
    parser_bar.add_argument('z', help='string')
    parser_bar.set_defaults(func=base_bar)

    # create the parser for the "sport" subcommand
    parser_sport = base_subparsers.add_parser('sport', help='sport help')
    sport_arg = parser_sport.add_argument('sport', help='Enter name of a sport', choices=sport_item_strs)

    @cmd2.with_argparser(base_parser, with_unknown_args=True)
    def do_base(self, args) -> None:
        """Base command help"""
        func = getattr(args, 'func', None)
        if func is not None:
            # Call whatever subcommand function was selected
            func(self, args)
        else:
            # No subcommand was provided, so call help
            self.do_help('base')


@pytest.fixture
def scu_app():
    """Declare test fixture for with_argparser decorator"""
    return SubcommandsWithUnknownExample()


def test_subcmd_with_unknown_completion_single_end(scu_app) -> None:
    text = 'f'
    line = f'base {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, scu_app)

    print(f'first_match: {first_match}')

    # It is at end of line, so extra space is present
    assert first_match is not None
    assert scu_app.completion_matches == ['foo ']


def test_subcmd_with_unknown_completion_multiple(scu_app) -> None:
    text = ''
    line = f'base {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, scu_app)
    assert first_match is not None
    assert scu_app.completion_matches == ['bar', 'foo', 'sport']


def test_subcmd_with_unknown_completion_nomatch(scu_app) -> None:
    text = 'z'
    line = f'base {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, scu_app)
    assert first_match is None


def test_help_subcommand_completion_single_scu(scu_app) -> None:
    text = 'base'
    line = f'help {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, scu_app)

    # It is at end of line, so extra space is present
    assert first_match is not None
    assert scu_app.completion_matches == ['base ']


def test_help_subcommand_completion_multiple_scu(scu_app) -> None:
    text = ''
    line = f'help base {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, scu_app)
    assert first_match is not None
    assert scu_app.completion_matches == ['bar', 'foo', 'sport']


def test_help_subcommand_completion_with_flags_before_command(scu_app) -> None:
    text = ''
    line = f'help -h -v base {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, scu_app)
    assert first_match is not None
    assert scu_app.completion_matches == ['bar', 'foo', 'sport']


def test_complete_help_subcommands_with_blank_command(scu_app) -> None:
    text = ''
    line = f'help "" {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, scu_app)
    assert first_match is None
    assert not scu_app.completion_matches


def test_help_subcommand_completion_nomatch_scu(scu_app) -> None:
    text = 'z'
    line = f'help base {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, scu_app)
    assert first_match is None


def test_subcommand_tab_completion_scu(scu_app) -> None:
    # This makes sure the correct completer for the sport subcommand is called
    text = 'Foot'
    line = f'base sport {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, scu_app)

    # It is at end of line, so extra space is present
    assert first_match is not None
    assert scu_app.completion_matches == ['Football ']


def test_subcommand_tab_completion_with_no_completer_scu(scu_app) -> None:
    # This tests what happens when a subcommand has no completer
    # In this case, the foo subcommand has no completer defined
    text = 'Foot'
    line = f'base foo {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, scu_app)
    assert first_match is None


def test_subcommand_tab_completion_space_in_text_scu(scu_app) -> None:
    text = 'B'
    line = f'base sport "Space {text}'
    endidx = len(line)
    begidx = endidx - len(text)

    first_match = complete_tester(text, line, begidx, endidx, scu_app)

    assert first_match is not None
    assert scu_app.completion_matches == ['Ball" ']
    assert scu_app.display_matches == ['Space Ball']



#!/usr/bin/env python
"""A simple example demonstrating support for unicode command names."""

import math

import cmd2


class UnicodeApp(cmd2.Cmd):
    """Example cmd2 application with unicode command names."""

    def __init__(self) -> None:
        super().__init__()
        self.intro = 'Welcome the Unicode example app. Note the full Unicode support:  😇 💩'

    def do_𝛑print(self, _) -> None:  # noqa: PLC2401
        """This command prints 𝛑 to 5 decimal places."""
        self.poutput(f"𝛑 = {math.pi:.6}")

    def do_你好(self, arg) -> None:  # noqa: N802, PLC2401
        """This command says hello in Chinese (Mandarin)."""
        self.poutput("你好 " + arg)


if __name__ == '__main__':
    app = UnicodeApp()
    app.cmdloop()



#!/usr/bin/env python
"""A sample application for cmd2 demonstrating how to use hooks.

This application shows how to use postparsing hooks to allow case insensitive
command names, abbreviated commands, as well as allowing numeric arguments to
follow a command without any intervening whitespace.

"""

import re

import cmd2


class CmdLineApp(cmd2.Cmd):
    """Example cmd2 application demonstrating the use of hooks.

    This simple application has one command, `list` which generates a list
    of 10 numbers. This command takes one optional argument, which is the
    number to start on.

    We have three postparsing hooks, which allow the user to enter:

        (Cmd) list 5
        (Cmd) L 5
        (Cmd) l 5
        (Cmd) L5
        (Cmd) LI5

    and have them all treated as valid input which prints a list of 10 numbers
    starting with the number 5.

    We also add a postcommand hook, which updates the shell prompt to show the
    raw contents of the Statement after the postparsing hooks are finished. To
    use this hook, run `(Cmd) set debug True`. All of the above variations of
    the list command should produce the same raw content.

    """

    # Setting this true makes it run a shell command if a cmd2/cmd command doesn't exist
    # default_to_shell = True  # noqa: ERA001
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        # register four hooks
        self.register_postparsing_hook(self.add_whitespace_hook)
        self.register_postparsing_hook(self.downcase_hook)
        self.register_postparsing_hook(self.abbrev_hook)
        self.register_postcmd_hook(self.proof_hook)

    def add_whitespace_hook(self, data: cmd2.plugin.PostparsingData) -> cmd2.plugin.PostparsingData:
        """A hook to split alphabetic command names immediately followed by a number.

        l24 -> l 24
        list24 -> list 24
        list 24 -> list 24

        """
        command = data.statement.command
        # regular expression with looks for:
        #  ^ - the beginning of the string
        # ([^\s\d]+) - one or more non-whitespace non-digit characters, set as capture group 1
        # (\d+) - one or more digit characters, set as capture group 2
        command_pattern = re.compile(r'^([^\s\d]+)(\d+)')
        match = command_pattern.search(command)
        if match:
            command = match.group(1)
            first_arg = match.group(2)
            rest_args = data.statement.args
            post_command = data.statement.post_command
            data.statement = self.statement_parser.parse(f'{command} {first_arg} {rest_args} {post_command}')
        return data

    def downcase_hook(self, data: cmd2.plugin.PostparsingData) -> cmd2.plugin.PostparsingData:
        """A hook to make uppercase commands lowercase."""
        command = data.statement.command.lower()
        args = data.statement.args
        post_command = data.statement.post_command
        data.statement = self.statement_parser.parse(f'{command} {args} {post_command}')
        return data

    def abbrev_hook(self, data: cmd2.plugin.PostparsingData) -> cmd2.plugin.PostparsingData:
        """Accept unique abbreviated commands."""
        func = self.cmd_func(data.statement.command)
        if func is None:
            # check if the entered command might be an abbreviation
            possible_cmds = [cmd for cmd in self.get_all_commands() if cmd.startswith(data.statement.command)]
            if len(possible_cmds) == 1:
                raw = data.statement.raw.replace(data.statement.command, possible_cmds[0], 1)
                data.statement = self.statement_parser.parse(raw)
        return data

    def proof_hook(self, data: cmd2.plugin.PostcommandData) -> cmd2.plugin.PostcommandData:
        """Update the shell prompt with the new raw statement after postparsing hooks are finished."""
        if self.debug:
            self.prompt = f'({data.statement.raw})'
        return data

    @cmd2.with_argument_list
    def do_list(self, arglist: list[str]) -> None:
        """Generate a list of 10 numbers."""
        if arglist:
            first = arglist[0]
            try:
                first = int(first)
            except ValueError:
                first = 1
        else:
            first = 1
        last = first + 10

        for x in range(first, last):
            self.poutput(str(x))


if __name__ == '__main__':
    import sys

    c = CmdLineApp()
    sys.exit(c.cmdloop())



#!/usr/bin/env python
"""This is intended to be a completely bare-bones cmd2 application suitable for rapid testing and debugging."""

from cmd2 import (
    cmd2,
)

if __name__ == '__main__':
    import sys

    # If run as the main application, simply start a bare-bones cmd2 application with only built-in functionality.
    # Enable commands to support interactive Python and IPython shells.
    app = cmd2.Cmd(include_py=True, include_ipy=True, persistent_history_file='cmd2_history.dat')
    app.self_in_py = True  # Enable access to "self" within the py command
    app.debug = True  # Show traceback if/when an exception occurs
    sys.exit(app.cmdloop())



"""Unit testing for cmd2/rich_utils.py module"""

import pytest
import rich.box
from rich.console import Console
from rich.style import Style
from rich.table import Table
from rich.text import Text

from cmd2 import (
    Cmd2Style,
    Color,
)
from cmd2 import rich_utils as ru


def test_cmd2_base_console() -> None:
    # Test the keyword arguments which are not allowed.
    with pytest.raises(TypeError) as excinfo:
        ru.Cmd2BaseConsole(force_terminal=True)
    assert 'force_terminal' in str(excinfo.value)

    with pytest.raises(TypeError) as excinfo:
        ru.Cmd2BaseConsole(force_interactive=True)
    assert 'force_interactive' in str(excinfo.value)

    with pytest.raises(TypeError) as excinfo:
        ru.Cmd2BaseConsole(theme=None)
    assert 'theme' in str(excinfo.value)


def test_indented_text() -> None:
    console = Console(width=20)

    # With an indention of 10, text will be evenly split across two lines.
    text = "A" * 20
    level = 10
    indented_text = ru.indent(text, level)

    with console.capture() as capture:
        console.print(indented_text)
    result = capture.get().splitlines()

    padding = " " * level
    expected_line = padding + ("A" * 10)
    assert result[0] == expected_line
    assert result[1] == expected_line


def test_indented_table() -> None:
    console = Console()

    level = 2
    table = Table("Column", box=rich.box.ASCII)
    table.add_row("Some Data")
    indented_table = ru.indent(table, level)

    with console.capture() as capture:
        console.print(indented_table)
    result = capture.get().splitlines()

    padding = " " * level
    assert result[0].startswith(padding + "+-----------+")
    assert result[1].startswith(padding + "| Column    |")
    assert result[2].startswith(padding + "|-----------|")
    assert result[3].startswith(padding + "| Some Data |")
    assert result[4].startswith(padding + "+-----------+")


@pytest.mark.parametrize(
    ('rich_text', 'string'),
    [
        (Text("Hello"), "Hello"),
        (Text("Hello\n"), "Hello\n"),
        (Text("Hello", style="blue"), "\x1b[34mHello\x1b[0m"),
    ],
)
def test_rich_text_to_string(rich_text: Text, string: str) -> None:
    assert ru.rich_text_to_string(rich_text) == string


def test_set_theme() -> None:
    # Save a cmd2, rich-argparse, and rich-specific style.
    cmd2_style_key = Cmd2Style.ERROR
    argparse_style_key = "argparse.args"
    rich_style_key = "inspect.attr"

    orig_cmd2_style = ru.APP_THEME.styles[cmd2_style_key]
    orig_argparse_style = ru.APP_THEME.styles[argparse_style_key]
    orig_rich_style = ru.APP_THEME.styles[rich_style_key]

    # Overwrite these styles by setting a new theme.
    theme = {
        cmd2_style_key: Style(color=Color.CYAN),
        argparse_style_key: Style(color=Color.AQUAMARINE3, underline=True),
        rich_style_key: Style(color=Color.DARK_GOLDENROD, bold=True),
    }
    ru.set_theme(theme)

    # Verify theme styles have changed to our custom values.
    assert ru.APP_THEME.styles[cmd2_style_key] != orig_cmd2_style
    assert ru.APP_THEME.styles[cmd2_style_key] == theme[cmd2_style_key]

    assert ru.APP_THEME.styles[argparse_style_key] != orig_argparse_style
    assert ru.APP_THEME.styles[argparse_style_key] == theme[argparse_style_key]

    assert ru.APP_THEME.styles[rich_style_key] != orig_rich_style
    assert ru.APP_THEME.styles[rich_style_key] == theme[rich_style_key]


def test_from_ansi_wrapper() -> None:
    # Check if we are still patching Text.from_ansi(). If this check fails, then Rich
    # has fixed the bug. Therefore, we can remove this test function and ru._from_ansi_wrapper.
    assert Text.from_ansi.__func__ is ru._from_ansi_wrapper.__func__  # type: ignore[attr-defined]

    # Line breaks recognized by str.splitlines().
    # Source: https://docs.python.org/3/library/stdtypes.html#str.splitlines
    line_breaks = {
        "\n",  # Line Feed
        "\r",  # Carriage Return
        "\r\n",  # Carriage Return + Line Feed
        "\v",  # Vertical Tab
        "\f",  # Form Feed
        "\x1c",  # File Separator
        "\x1d",  # Group Separator
        "\x1e",  # Record Separator
        "\x85",  # Next Line (NEL)
        "\u2028",  # Line Separator
        "\u2029",  # Paragraph Separator
    }

    # Test all line breaks
    for lb in line_breaks:
        input_string = f"Text{lb}"
        expected_output = input_string.replace(lb, "\n")
        assert Text.from_ansi(input_string).plain == expected_output

    # Test string without trailing line break
    input_string = "No trailing\nline break"
    assert Text.from_ansi(input_string).plain == input_string

    # Test empty string
    input_string = ""
    assert Text.from_ansi(input_string).plain == input_string



"""A simple example demonstrating a loadable command set."""

from cmd2 import (
    CommandSet,
    Statement,
    with_default_category,
)


@with_default_category('Custom Init')
class CustomInitCommandSet(CommandSet):
    def __init__(self, arg1, arg2) -> None:
        super().__init__()

        self._arg1 = arg1
        self._arg2 = arg2

    def do_show_arg1(self, _: Statement) -> None:
        self._cmd.poutput('Arg1: ' + self._arg1)

    def do_show_arg2(self, _: Statement) -> None:
        self._cmd.poutput('Arg2: ' + self._arg2)



# Tests echo argument to app()
app.cmd_echo = False

# echo defaults to current setting which is False, so this help text should not be echoed to pytest's stdout
app('help alias')

# pytest's stdout should have this help text written to it
app('help edit', echo=True)



#!/usr/bin/env python
"""A sample application for cmd2.

This example is very similar to transcript_example.py, but had additional
code in main() that shows how to accept a command from
the command line at invocation:

$ python cmd_as_argument.py speak -p hello there


"""

import argparse
import random

import cmd2


class CmdLineApp(cmd2.Cmd):
    """Example cmd2 application."""

    # Setting this true makes it run a shell command if a cmd2/cmd command doesn't exist
    # default_to_shell = True  # noqa: ERA001
    MUMBLES = ('like', '...', 'um', 'er', 'hmmm', 'ahh')
    MUMBLE_FIRST = ('so', 'like', 'well')
    MUMBLE_LAST = ('right?',)

    def __init__(self) -> None:
        shortcuts = dict(cmd2.DEFAULT_SHORTCUTS)
        shortcuts.update({'&': 'speak'})
        # Set include_ipy to True to enable the "ipy" command which runs an interactive IPython shell
        super().__init__(allow_cli_args=False, include_ipy=True, multiline_commands=['orate'], shortcuts=shortcuts)

        self.self_in_py = True
        self.maxrepeats = 3
        # Make maxrepeats settable at runtime
        self.add_settable(cmd2.Settable('maxrepeats', int, 'max repetitions for speak command', self))

    speak_parser = cmd2.Cmd2ArgumentParser()
    speak_parser.add_argument('-p', '--piglatin', action='store_true', help='atinLay')
    speak_parser.add_argument('-s', '--shout', action='store_true', help='N00B EMULATION MODE')
    speak_parser.add_argument('-r', '--repeat', type=int, help='output [n] times')
    speak_parser.add_argument('words', nargs='+', help='words to say')

    @cmd2.with_argparser(speak_parser)
    def do_speak(self, args) -> None:
        """Repeats what you tell me to."""
        words = []
        for word in args.words:
            if args.piglatin:
                word = f'{word[1:]}{word[0]}ay'
            if args.shout:
                word = word.upper()
            words.append(word)
        repetitions = args.repeat or 1
        for _ in range(min(repetitions, self.maxrepeats)):
            # .poutput handles newlines, and accommodates output redirection too
            self.poutput(' '.join(words))

    do_say = do_speak  # now "say" is a synonym for "speak"
    do_orate = do_speak  # another synonym, but this one takes multi-line input

    mumble_parser = cmd2.Cmd2ArgumentParser()
    mumble_parser.add_argument('-r', '--repeat', type=int, help='how many times to repeat')
    mumble_parser.add_argument('words', nargs='+', help='words to say')

    @cmd2.with_argparser(mumble_parser)
    def do_mumble(self, args) -> None:
        """Mumbles what you tell me to."""
        repetitions = args.repeat or 1
        for _ in range(min(repetitions, self.maxrepeats)):
            output = []
            if random.random() < 0.33:
                output.append(random.choice(self.MUMBLE_FIRST))
            for word in args.words:
                if random.random() < 0.40:
                    output.append(random.choice(self.MUMBLES))
                output.append(word)
            if random.random() < 0.25:
                output.append(random.choice(self.MUMBLE_LAST))
            self.poutput(' '.join(output))


def main(argv=None):
    """Run when invoked from the operating system shell."""
    parser = cmd2.Cmd2ArgumentParser(description='Commands as arguments')
    command_help = 'optional command to run, if no command given, enter an interactive shell'
    parser.add_argument('command', nargs='?', help=command_help)
    arg_help = 'optional arguments for command'
    parser.add_argument('command_args', nargs=argparse.REMAINDER, help=arg_help)

    args = parser.parse_args(argv)

    c = CmdLineApp()

    sys_exit_code = 0
    if args.command:
        # we have a command, run it and then exit
        c.onecmd_plus_hooks('{} {}'.format(args.command, ' '.join(args.command_args)))
    else:
        # we have no command, drop into interactive mode
        sys_exit_code = c.cmdloop()

    return sys_exit_code


if __name__ == '__main__':
    import sys

    sys.exit(main())



#!/usr/bin/env python3
"""A comprehensive example demonstrating various aspects of using `argparse` for command argument processing.

Demonstrates basic usage of the `cmd2.with_argparser` decorator for passing a `cmd2.Cmd2ArgumentParser` to a `do_*` command
method. The `fsize` and `pow` commands demonstrate various different types of arguments, actions, choices, and completers that
can be used.

The `print_args` and `print_unknown` commands display how argparse arguments are passed to commands in the cases that unknown
arguments are not captured and are captured, respectively.

The `base` and `alternate` commands show an easy way for a single command to have many subcommands, each of which take
different arguments and provides separate contextual help.

Lastly, this example shows how you can also use `argparse` to parse command-line arguments when launching a cmd2 application.
"""

import argparse
import os

import cmd2
from cmd2.string_utils import stylize

# Command categories
ARGPARSE_USAGE = 'Argparse Basic Usage'
ARGPARSE_PRINTING = 'Argparse Printing'
ARGPARSE_SUBCOMMANDS = 'Argparse Subcommands'


class ArgparsingApp(cmd2.Cmd):
    def __init__(self, color: str) -> None:
        """Cmd2 application for demonstrating the use of argparse for command argument parsing."""
        super().__init__(include_ipy=True)
        self.intro = stylize(
            'cmd2 has awesome decorators to make it easy to use Argparse to parse command arguments', style=color
        )

    ## ------ Basic examples of using argparse for command argument parsing -----

    # do_fsize parser
    fsize_parser = cmd2.Cmd2ArgumentParser(description='Obtain the size of a file')
    fsize_parser.add_argument('-c', '--comma', action='store_true', help='add comma for thousands separator')
    fsize_parser.add_argument('-u', '--unit', choices=['MB', 'KB'], help='unit to display size in')
    fsize_parser.add_argument('file_path', help='path of file', completer=cmd2.Cmd.path_complete)

    @cmd2.with_argparser(fsize_parser)
    @cmd2.with_category(ARGPARSE_USAGE)
    def do_fsize(self, args: argparse.Namespace) -> None:
        """Obtain the size of a file."""
        expanded_path = os.path.expanduser(args.file_path)

        try:
            size = os.path.getsize(expanded_path)
        except OSError as ex:
            self.perror(f"Error retrieving size: {ex}")
            return

        if args.unit == 'KB':
            size //= 1024
        elif args.unit == 'MB':
            size //= 1024 * 1024
        else:
            args.unit = 'bytes'
        size = round(size, 2)

        size_str = f'{size:,}' if args.comma else f'{size}'
        self.poutput(f'{size_str} {args.unit}')

    # do_pow parser
    pow_parser = cmd2.Cmd2ArgumentParser()
    pow_parser.add_argument('base', type=int)
    pow_parser.add_argument('exponent', type=int, choices=range(-5, 6))

    @cmd2.with_argparser(pow_parser)
    @cmd2.with_category(ARGPARSE_USAGE)
    def do_pow(self, args: argparse.Namespace) -> None:
        """Raise an integer to a small integer exponent, either positive or negative.

        :param args: argparse arguments
        """
        self.poutput(f'{args.base} ** {args.exponent} == {args.base**args.exponent}')

    ## ------ Examples displaying how argparse arguments are passed to commands by printing them out -----

    argprint_parser = cmd2.Cmd2ArgumentParser()
    argprint_parser.add_argument('-p', '--piglatin', action='store_true', help='atinLay')
    argprint_parser.add_argument('-s', '--shout', action='store_true', help='N00B EMULATION MODE')
    argprint_parser.add_argument('-r', '--repeat', type=int, help='output [n] times')
    argprint_parser.add_argument('words', nargs='+', help='words to print')

    @cmd2.with_argparser(argprint_parser)
    @cmd2.with_category(ARGPARSE_PRINTING)
    def do_print_args(self, args: argparse.Namespace) -> None:
        """Print the arpgarse argument list this command was called with."""
        self.poutput(f'print_args was called with the following\n\targuments: {args!r}')

    unknownprint_parser = cmd2.Cmd2ArgumentParser()
    unknownprint_parser.add_argument('-p', '--piglatin', action='store_true', help='atinLay')
    unknownprint_parser.add_argument('-s', '--shout', action='store_true', help='N00B EMULATION MODE')
    unknownprint_parser.add_argument('-r', '--repeat', type=int, help='output [n] times')

    @cmd2.with_argparser(unknownprint_parser, with_unknown_args=True)
    @cmd2.with_category(ARGPARSE_PRINTING)
    def do_print_unknown(self, args: argparse.Namespace, unknown: list[str]) -> None:
        """Print the arpgarse argument list this command was called with, including unknown arguments."""
        self.poutput(f'print_unknown was called with the following arguments\n\tknown: {args!r}\n\tunknown: {unknown}')

    ## ------ Examples demonstrating how to use argparse subcommands -----

    # create the top-level parser for the base command
    calculate_parser = cmd2.Cmd2ArgumentParser(description="Perform simple mathematical calculations.")
    calculate_subparsers = calculate_parser.add_subparsers(title='operation', help='Available operations', required=True)

    # create the parser for the "add" subcommand
    add_description = "Add two numbers"
    add_parser = cmd2.Cmd2ArgumentParser("add", description=add_description)
    add_parser.add_argument('num1', type=int, help='The first number')
    add_parser.add_argument('num2', type=int, help='The second number')

    # create the parser for the "add" subcommand
    subtract_description = "Subtract two numbers"
    subtract_parser = cmd2.Cmd2ArgumentParser("subtract", description=subtract_description)
    subtract_parser.add_argument('num1', type=int, help='The first number')
    subtract_parser.add_argument('num2', type=int, help='The second number')

    # subcommand functions for the calculate command
    @cmd2.as_subcommand_to('calculate', 'add', add_parser, help=add_description.lower())
    def add(self, args: argparse.Namespace) -> None:
        """add subcommand of calculate command."""
        result = args.num1 + args.num2
        self.poutput(f"{args.num1} + {args.num2} = {result}")

    @cmd2.as_subcommand_to('calculate', 'subtract', subtract_parser, help=subtract_description.lower())
    def subtract(self, args: argparse.Namespace) -> None:
        """subtract subcommand of calculate command."""
        result = args.num1 - args.num2
        self.poutput(f"{args.num1} - {args.num2} = {result}")

    @cmd2.with_argparser(calculate_parser)
    @cmd2.with_category(ARGPARSE_SUBCOMMANDS)
    def do_calculate(self, args: argparse.Namespace) -> None:
        """Calculate a simple mathematical operation on two integers."""
        handler = args.cmd2_handler.get()
        handler(args)


if __name__ == '__main__':
    import sys

    from cmd2.colors import Color

    # You can do your custom Argparse parsing here to meet your application's needs
    parser = cmd2.Cmd2ArgumentParser(description='Process the arguments however you like.')

    # Add an argument which we will pass to the app to change some behavior
    parser.add_argument(
        '-c',
        '--color',
        choices=[Color.RED, Color.ORANGE1, Color.YELLOW, Color.GREEN, Color.BLUE, Color.PURPLE, Color.VIOLET, Color.WHITE],
        help='Color of intro text',
    )

    # Parse the arguments
    args, unknown_args = parser.parse_known_args()

    color = Color.WHITE
    if args.color:
        color = args.color

    # Perform surgery on sys.argv to remove the arguments which have already been processed by argparse
    sys.argv = sys.argv[:1] + unknown_args

    # Instantiate your cmd2 application
    app = ArgparsingApp(color)

    # And run your cmd2 application
    sys.exit(app.cmdloop())



#!/usr/bin/env python3
"""Example revolving around the CommandSet feature for modularizing commands.

It attempts to cover basic usage as well as more complex usage including dynamic loading and unloading of CommandSets, using
CommandSets to add subcommands, as well as how to categorize command in CommandSets. Here we have kept the implementation for
most commands trivial because the intent is to focus on the CommandSet feature set.

The `AutoLoadCommandSet` is a basic command set which is loaded automatically at application startup and stays loaded until
application exit. Ths is the simplest case of simply modularizing command definitions to different classes and/or files.

The `LoadableFruits` and `LoadableVegetables` CommandSets are dynamically loadable and un-loadable at runtime using the `load`
and `unload` commands. This demonstrates the ability to load and unload CommandSets based on application state. Each of these
also loads a subcommand of the `cut` command.
"""

import argparse

import cmd2
from cmd2 import (
    CommandSet,
    with_argparser,
    with_category,
    with_default_category,
)

COMMANDSET_BASIC = "Basic CommandSet"
COMMANDSET_DYNAMIC = "Dynamic CommandSet"
COMMANDSET_LOAD_UNLOAD = "Loading and Unloading CommandSets"
COMMANDSET_SUBCOMMAND = "Subcommands with CommandSet"


@with_default_category(COMMANDSET_BASIC)
class AutoLoadCommandSet(CommandSet):
    def __init__(self) -> None:
        """CommandSet class for auto-loading commands at startup."""
        super().__init__()

    def do_hello(self, _: cmd2.Statement) -> None:
        """Print hello."""
        self._cmd.poutput('Hello')

    def do_world(self, _: cmd2.Statement) -> None:
        """Print World."""
        self._cmd.poutput('World')


@with_default_category(COMMANDSET_DYNAMIC)
class LoadableFruits(CommandSet):
    def __init__(self) -> None:
        """CommandSet class for dynamically loading commands related to fruits."""
        super().__init__()

    def do_apple(self, _: cmd2.Statement) -> None:
        """Print Apple."""
        self._cmd.poutput('Apple')

    def do_banana(self, _: cmd2.Statement) -> None:
        """Print Banana."""
        self._cmd.poutput('Banana')

    banana_description = "Cut a banana"
    banana_parser = cmd2.Cmd2ArgumentParser(description=banana_description)
    banana_parser.add_argument('direction', choices=['discs', 'lengthwise'])

    @cmd2.as_subcommand_to('cut', 'banana', banana_parser, help=banana_description.lower())
    def cut_banana(self, ns: argparse.Namespace) -> None:
        """Cut banana."""
        self._cmd.poutput('cutting banana: ' + ns.direction)


@with_default_category(COMMANDSET_DYNAMIC)
class LoadableVegetables(CommandSet):
    def __init__(self) -> None:
        """CommandSet class for dynamically loading commands related to vegetables."""
        super().__init__()

    def do_arugula(self, _: cmd2.Statement) -> None:
        "Print Arguula."
        self._cmd.poutput('Arugula')

    def do_bokchoy(self, _: cmd2.Statement) -> None:
        """Print Bok Choy."""
        self._cmd.poutput('Bok Choy')

    bokchoy_description = "Cut some bokchoy"
    bokchoy_parser = cmd2.Cmd2ArgumentParser(description=bokchoy_description)
    bokchoy_parser.add_argument('style', choices=['quartered', 'diced'])

    @cmd2.as_subcommand_to('cut', 'bokchoy', bokchoy_parser, help=bokchoy_description.lower())
    def cut_bokchoy(self, ns: argparse.Namespace) -> None:
        """Cut bokchoy."""
        self._cmd.poutput('Bok Choy: ' + ns.style)


class CommandSetApp(cmd2.Cmd):
    """CommandSets are automatically loaded. Nothing needs to be done."""

    def __init__(self) -> None:
        """Cmd2 application for demonstrating the CommandSet features."""
        # This prevents all CommandSets from auto-loading, which is necessary if you don't want some to load at startup
        super().__init__(auto_load_commands=False)

        self.register_command_set(AutoLoadCommandSet())

        # Store the dyanmic CommandSet classes for ease of loading and unloading
        self._fruits = LoadableFruits()
        self._vegetables = LoadableVegetables()

        self.intro = 'The CommandSet feature allows defining commands in multiple files and the dynamic load/unload at runtime'

    load_parser = cmd2.Cmd2ArgumentParser()
    load_parser.add_argument('cmds', choices=['fruits', 'vegetables'])

    @with_argparser(load_parser)
    @with_category(COMMANDSET_LOAD_UNLOAD)
    def do_load(self, ns: argparse.Namespace) -> None:
        """Load a CommandSet at runtime."""
        if ns.cmds == 'fruits':
            try:
                self.register_command_set(self._fruits)
                self.poutput('Fruits loaded')
            except ValueError:
                self.poutput('Fruits already loaded')

        if ns.cmds == 'vegetables':
            try:
                self.register_command_set(self._vegetables)
                self.poutput('Vegetables loaded')
            except ValueError:
                self.poutput('Vegetables already loaded')

    @with_argparser(load_parser)
    @with_category(COMMANDSET_LOAD_UNLOAD)
    def do_unload(self, ns: argparse.Namespace) -> None:
        """Unload a CommandSet at runtime."""
        if ns.cmds == 'fruits':
            self.unregister_command_set(self._fruits)
            self.poutput('Fruits unloaded')

        if ns.cmds == 'vegetables':
            self.unregister_command_set(self._vegetables)
            self.poutput('Vegetables unloaded')

    cut_parser = cmd2.Cmd2ArgumentParser()
    cut_subparsers = cut_parser.add_subparsers(title='item', help='item to cut')

    @with_argparser(cut_parser)
    @with_category(COMMANDSET_SUBCOMMAND)
    def do_cut(self, ns: argparse.Namespace) -> None:
        """Intended to be used with dyanmically loaded subcommands specifically."""
        handler = ns.cmd2_handler.get()
        if handler is not None:
            handler(ns)
        else:
            # No subcommand was provided, so call help
            self.poutput('This command does nothing without sub-parsers registered')
            self.do_help('cut')


if __name__ == '__main__':
    app = CommandSetApp()
    app.cmdloop()



"""Unit/functional testing for run_pytest in cmd2"""

import builtins
import os
from unittest import (
    mock,
)

import pytest

from cmd2.string_utils import quote

from .conftest import (
    odd_file_names,
    run_cmd,
)


def test_run_pyscript(base_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)
    python_script = os.path.join(test_dir, 'script.py')
    expected = 'This is a python script running ...'

    out, _err = run_cmd(base_app, f"run_pyscript {python_script}")
    assert expected in out
    assert base_app.last_result is True


def test_run_pyscript_recursive_not_allowed(base_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)
    python_script = os.path.join(test_dir, 'pyscript', 'recursive.py')
    expected = 'Recursively entering interactive Python shells is not allowed'

    _out, err = run_cmd(base_app, f"run_pyscript {python_script}")
    assert err[0] == expected
    assert base_app.last_result is False


def test_run_pyscript_with_nonexist_file(base_app) -> None:
    python_script = 'does_not_exist.py'
    _out, err = run_cmd(base_app, f"run_pyscript {python_script}")
    assert "Error reading script file" in err[0]
    assert base_app.last_result is False


def test_run_pyscript_with_non_python_file(base_app, request) -> None:
    m = mock.MagicMock(name='input', return_value='2')
    builtins.input = m

    test_dir = os.path.dirname(request.module.__file__)
    filename = os.path.join(test_dir, 'scripts', 'help.txt')
    _out, err = run_cmd(base_app, f'run_pyscript {filename}')
    assert "does not have a .py extension" in err[0]
    assert base_app.last_result is False


@pytest.mark.parametrize('python_script', odd_file_names)
def test_run_pyscript_with_odd_file_names(base_app, python_script) -> None:
    """Pass in file names with various patterns. Since these files don't exist, we will rely
    on the error text to make sure the file names were processed correctly.
    """
    # Mock input to get us passed the warning about not ending in .py
    input_mock = mock.MagicMock(name='input', return_value='1')
    builtins.input = input_mock

    _out, err = run_cmd(base_app, f"run_pyscript {quote(python_script)}")
    err = ''.join(err)
    assert f"Error reading script file '{python_script}'" in err
    assert base_app.last_result is False


def test_run_pyscript_with_exception(base_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)
    python_script = os.path.join(test_dir, 'pyscript', 'raises_exception.py')
    _out, err = run_cmd(base_app, f"run_pyscript {python_script}")
    assert err[0].startswith('Traceback')
    assert "TypeError: unsupported operand type(s) for +: 'int' and 'str'" in err[-1]
    assert base_app.last_result is True


def test_run_pyscript_requires_an_argument(base_app) -> None:
    _out, err = run_cmd(base_app, "run_pyscript")
    assert "the following arguments are required: script_path" in err[1]
    assert base_app.last_result is None


def test_run_pyscript_help(base_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)
    python_script = os.path.join(test_dir, 'pyscript', 'help.py')
    out1, _err1 = run_cmd(base_app, 'help')
    out2, _err2 = run_cmd(base_app, f'run_pyscript {python_script}')
    assert out1
    assert out1 == out2


def test_scripts_add_to_history(base_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)
    python_script = os.path.join(test_dir, 'pyscript', 'help.py')
    command = f'run_pyscript {python_script}'

    # Add to history
    base_app.scripts_add_to_history = True
    base_app.history.clear()
    run_cmd(base_app, command)
    assert len(base_app.history) == 2
    assert base_app.history.get(1).raw == command
    assert base_app.history.get(2).raw == 'help'

    # Do not add to history
    base_app.scripts_add_to_history = False
    base_app.history.clear()
    run_cmd(base_app, command)
    assert len(base_app.history) == 1
    assert base_app.history.get(1).raw == command


def test_run_pyscript_dir(base_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)
    python_script = os.path.join(test_dir, 'pyscript', 'pyscript_dir.py')

    out, _err = run_cmd(base_app, f'run_pyscript {python_script}')
    assert out[0] == "['cmd_echo']"


def test_run_pyscript_capture(base_app, request) -> None:
    base_app.self_in_py = True
    test_dir = os.path.dirname(request.module.__file__)
    python_script = os.path.join(test_dir, 'pyscript', 'stdout_capture.py')
    out, _err = run_cmd(base_app, f'run_pyscript {python_script}')

    assert out[0] == "print"
    assert out[1] == "poutput"


def test_run_pyscript_capture_custom_stdout(base_app, request) -> None:
    """sys.stdout will not be captured if it's different than self.stdout."""
    import io

    base_app.stdout = io.StringIO()

    base_app.self_in_py = True
    test_dir = os.path.dirname(request.module.__file__)
    python_script = os.path.join(test_dir, 'pyscript', 'stdout_capture.py')
    out, _err = run_cmd(base_app, f'run_pyscript {python_script}')

    assert "print" not in out
    assert out[0] == "poutput"


def test_run_pyscript_stop(base_app, request) -> None:
    # Verify onecmd_plus_hooks() returns True if any commands in a pyscript return True for stop
    test_dir = os.path.dirname(request.module.__file__)

    # help.py doesn't run any commands that return True for stop
    python_script = os.path.join(test_dir, 'pyscript', 'help.py')
    stop = base_app.onecmd_plus_hooks(f'run_pyscript {python_script}')
    assert not stop

    # stop.py runs the quit command which does return True for stop
    python_script = os.path.join(test_dir, 'pyscript', 'stop.py')
    stop = base_app.onecmd_plus_hooks(f'run_pyscript {python_script}')
    assert stop


def test_run_pyscript_environment(base_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)
    python_script = os.path.join(test_dir, 'pyscript', 'environment.py')
    out, _err = run_cmd(base_app, f'run_pyscript {python_script}')

    assert out[0] == "PASSED"


def test_run_pyscript_self_in_py(base_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)
    python_script = os.path.join(test_dir, 'pyscript', 'self_in_py.py')

    # Set self_in_py to True and make sure we see self
    base_app.self_in_py = True
    out, _err = run_cmd(base_app, f'run_pyscript {python_script}')
    assert 'I see self' in out[0]

    # Set self_in_py to False and make sure we can't see self
    base_app.self_in_py = False
    out, _err = run_cmd(base_app, f'run_pyscript {python_script}')
    assert 'I do not see self' in out[0]


def test_run_pyscript_py_locals(base_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)
    python_script = os.path.join(test_dir, 'pyscript', 'py_locals.py')

    # Make sure pyscripts can't edit Cmd.py_locals. It used to be that cmd2 was passing its py_locals
    # dictionary to the py environment instead of a shallow copy.
    base_app.py_locals['test_var'] = 5

    # Place an editable object in py_locals. Since we make a shallow copy of py_locals,
    # this object should be editable from the py environment.
    base_app.py_locals['my_list'] = []

    run_cmd(base_app, f'run_pyscript {python_script}')

    # test_var should still exist
    assert base_app.py_locals['test_var'] == 5

    # my_list should be edited
    assert base_app.py_locals['my_list'][0] == 2


def test_run_pyscript_app_echo(base_app, request) -> None:
    test_dir = os.path.dirname(request.module.__file__)
    python_script = os.path.join(test_dir, 'pyscript', 'echo.py')
    out, _err = run_cmd(base_app, f'run_pyscript {python_script}')

    # Only the edit help text should have been echoed to pytest's stdout
    assert out[0] == "Usage: edit [-h] [file_path]"



"""A simple example demonstrating a loadable command set."""

from cmd2 import (
    CommandSet,
    CompletionError,
    Statement,
    with_category,
    with_default_category,
)


@with_default_category('Basic Completion')
class BasicCompletionCommandSet(CommandSet):
    # List of strings used with completion functions
    food_item_strs = ('Pizza', 'Ham', 'Ham Sandwich', 'Potato')
    sport_item_strs = ('Bat', 'Basket', 'Basketball', 'Football', 'Space Ball')

    # This data is used to demonstrate delimiter_complete
    file_strs = (
        '/home/user/file.db',
        '/home/user/file space.db',
        '/home/user/another.db',
        '/home/other user/maps.db',
        '/home/other user/tests.db',
    )

    def do_flag_based(self, statement: Statement) -> None:
        """Tab completes arguments based on a preceding flag using flag_based_complete
        -f, --food [completes food items]
        -s, --sport [completes sports]
        -p, --path [completes local file system paths].
        """
        self._cmd.poutput(f"Args: {statement.args}")

    def complete_flag_based(self, text: str, line: str, begidx: int, endidx: int) -> list[str]:
        """Completion function for do_flag_based."""
        flag_dict = {
            # Tab complete food items after -f and --food flags in command line
            '-f': self.food_item_strs,
            '--food': self.food_item_strs,
            # Tab complete sport items after -s and --sport flags in command line
            '-s': self.sport_item_strs,
            '--sport': self.sport_item_strs,
            # Tab complete using path_complete function after -p and --path flags in command line
            '-p': self._cmd.path_complete,
            '--path': self._cmd.path_complete,
        }

        return self._cmd.flag_based_complete(text, line, begidx, endidx, flag_dict=flag_dict)

    def do_index_based(self, statement: Statement) -> None:
        """Tab completes first 3 arguments using index_based_complete."""
        self._cmd.poutput(f"Args: {statement.args}")

    def complete_index_based(self, text: str, line: str, begidx: int, endidx: int) -> list[str]:
        """Completion function for do_index_based."""
        index_dict = {
            1: self.food_item_strs,  # Tab complete food items at index 1 in command line
            2: self.sport_item_strs,  # Tab complete sport items at index 2 in command line
            3: self._cmd.path_complete,  # Tab complete using path_complete function at index 3 in command line
        }

        return self._cmd.index_based_complete(text, line, begidx, endidx, index_dict=index_dict)

    def do_delimiter_complete(self, statement: Statement) -> None:
        """Tab completes files from a list using delimiter_complete."""
        self._cmd.poutput(f"Args: {statement.args}")

    def complete_delimiter_complete(self, text: str, line: str, begidx: int, endidx: int) -> list[str]:
        return self._cmd.delimiter_complete(text, line, begidx, endidx, match_against=self.file_strs, delimiter='/')

    def do_raise_error(self, statement: Statement) -> None:
        """Demonstrates effect of raising CompletionError."""
        self._cmd.poutput(f"Args: {statement.args}")

    def complete_raise_error(self, _text: str, _line: str, _begidx: int, _endidx: int) -> list[str]:
        """CompletionErrors can be raised if an error occurs while tab completing.

        Example use cases
            - Reading a database to retrieve a tab completion data set failed
            - A previous command line argument that determines the data set being completed is invalid
        """
        raise CompletionError("This is how a CompletionError behaves")

    @with_category('Not Basic Completion')
    def do_custom_category(self, _statement: Statement) -> None:
        self._cmd.poutput('Demonstrates a command that bypasses the default category')



"""Unit testing for cmd2/string_utils.py module"""

from rich.style import Style

from cmd2 import Color
from cmd2 import rich_utils as ru
from cmd2 import string_utils as su

HELLO_WORLD = 'Hello, world!'


def test_align_blank() -> None:
    text = ''
    character = '-'
    width = 5
    aligned = su.align(text, "left", width=width, character=character)
    assert aligned == character * width


def test_align_wider_than_width() -> None:
    text = 'long text field'
    character = '-'
    width = 8
    aligned = su.align(text, "left", width=width, character=character)
    assert aligned == text[:width]


def test_align_term_width() -> None:
    text = 'foo'
    character = ' '

    term_width = ru.console_width()
    expected_padding = (term_width - su.str_width(text)) * character

    aligned = su.align(text, "left", character=character)
    assert aligned == text + expected_padding


def test_align_left() -> None:
    text = 'foo'
    character = '-'
    width = 5
    aligned = su.align_left(text, width=width, character=character)
    assert aligned == text + character * 2


def test_align_left_wide_text() -> None:
    text = '苹'
    character = '-'
    width = 4
    aligned = su.align_left(text, width=width, character=character)
    assert aligned == text + character * 2


def test_align_left_with_style() -> None:
    character = '-'

    styled_text = su.stylize('table', style=Color.BRIGHT_BLUE)
    width = 8

    aligned = su.align_left(styled_text, width=width, character=character)
    assert aligned == styled_text + character * 3


def test_align_center() -> None:
    text = 'foo'
    character = '-'
    width = 5
    aligned = su.align_center(text, width=width, character=character)
    assert aligned == character + text + character


def test_align_center_wide_text() -> None:
    text = '苹'
    character = '-'
    width = 4
    aligned = su.align_center(text, width=width, character=character)
    assert aligned == character + text + character


def test_align_center_with_style() -> None:
    character = '-'

    styled_text = su.stylize('table', style=Color.BRIGHT_BLUE)
    width = 8

    aligned = su.align_center(styled_text, width=width, character=character)
    assert aligned == character + styled_text + character * 2


def test_align_right() -> None:
    text = 'foo'
    character = '-'
    width = 5
    aligned = su.align_right(text, width=width, character=character)
    assert aligned == character * 2 + text


def test_align_right_wide_text() -> None:
    text = '苹'
    character = '-'
    width = 4
    aligned = su.align_right(text, width=width, character=character)
    assert aligned == character * 2 + text


def test_align_right_with_style() -> None:
    character = '-'

    styled_text = su.stylize('table', style=Color.BRIGHT_BLUE)
    width = 8

    aligned = su.align_right(styled_text, width=width, character=character)
    assert aligned == character * 3 + styled_text


def test_stylize() -> None:
    # Test string with no existing style
    style = Style(color=Color.GREEN, bgcolor=Color.BLUE, bold=True, underline=True)
    styled_str = su.stylize(HELLO_WORLD, style=style)
    assert styled_str == "\x1b[1;4;32;44mHello, world!\x1b[0m"

    # Add style to already-styled string
    updated_style = Style.combine([style, Style(strike=True)])
    restyled_string = su.stylize(styled_str, style=updated_style)
    assert restyled_string == "\x1b[1;4;9;32;44mHello, world!\x1b[0m"


def test_strip_style() -> None:
    base_str = HELLO_WORLD
    styled_str = su.stylize(base_str, style=Color.GREEN)
    assert base_str != styled_str
    assert base_str == su.strip_style(styled_str)


def test_str_width() -> None:
    # Include a full-width character
    base_str = HELLO_WORLD + "深"
    styled_str = su.stylize(base_str, style=Color.GREEN)
    expected_width = len(HELLO_WORLD) + 2
    assert su.str_width(base_str) == su.str_width(styled_str) == expected_width


def test_is_quoted_short() -> None:
    my_str = ''
    assert not su.is_quoted(my_str)
    your_str = '"'
    assert not su.is_quoted(your_str)


def test_is_quoted_yes() -> None:
    my_str = '"This is a test"'
    assert su.is_quoted(my_str)
    your_str = "'of the emergengy broadcast system'"
    assert su.is_quoted(your_str)


def test_is_quoted_no() -> None:
    my_str = '"This is a test'
    assert not su.is_quoted(my_str)
    your_str = "of the emergengy broadcast system'"
    assert not su.is_quoted(your_str)
    simple_str = "hello world"
    assert not su.is_quoted(simple_str)


def test_quote() -> None:
    my_str = "Hello World"
    assert su.quote(my_str) == '"' + my_str + '"'

    my_str = "'Hello World'"
    assert su.quote(my_str) == '"' + my_str + '"'

    my_str = '"Hello World"'
    assert su.quote(my_str) == "'" + my_str + "'"


def test_quote_if_needed_yes() -> None:
    my_str = "Hello World"
    assert su.quote_if_needed(my_str) == '"' + my_str + '"'
    your_str = '"foo" bar'
    assert su.quote_if_needed(your_str) == "'" + your_str + "'"


def test_quote_if_needed_no() -> None:
    my_str = "HelloWorld"
    assert su.quote_if_needed(my_str) == my_str
    your_str = "'Hello World'"
    assert su.quote_if_needed(your_str) == your_str


def test_strip_quotes_no_quotes() -> None:
    base_str = HELLO_WORLD
    stripped = su.strip_quotes(base_str)
    assert base_str == stripped


def test_strip_quotes_with_quotes() -> None:
    base_str = '"' + HELLO_WORLD + '"'
    stripped = su.strip_quotes(base_str)
    assert stripped == HELLO_WORLD


def test_unicode_normalization() -> None:
    s1 = 'café'
    s2 = 'cafe\u0301'
    assert s1 != s2
    assert su.norm_fold(s1) == su.norm_fold(s2)


def test_unicode_casefold() -> None:
    micro = 'µ'
    micro_cf = micro.casefold()
    assert micro != micro_cf
    assert su.norm_fold(micro) == su.norm_fold(micro_cf)
